grammar GDSLGrammar;

options {
    language=C;
    output=AST;
    ASTLabelType=pANTLR3_BASE_TREE;
    backtrack=true;
}
 
tokens {
    //All rows and columns
    ROWS        = 'rows';               
    COLUMNS     = 'cols';

    //Utilities
    DOTS        = '..';
    COMMA       = ',';
    BLANK       = '_';
    EQUALS      = '=';
    LEQ         = '<=';
    GEQ         = '>=';
    LESSER      = '<';
    GREATER     = '>';
    DEF         = ':=';
    DEFVAR      = 'def';

    AND         = 'and';
    WITH        = 'with';
    WITHOUT     = 'without';
    OF          = 'of';

    FOR         = 'for';
    IN          = 'in';
    DO          = 'do';
    SUM         = 'sum';
    END         = 'end';
    MATH        = 'mathroot';
    MATHFUNC    = 'mathfunc';
    INTVL       = 'intvlroot';
    OBJREST     = 'objrestroot';
    NUM         = 'numeric';
    SUMR        = 'sumroot';
    DOR         = 'doroot';
    NUMOF       = '#';
    LINEBREAK   = '\n';
    FOO         = '_foo';
}

@lexer::includes
{
    #define _empty NULL
}

@parser::includes
{
    #define _empty NULL
    #include "antlrparseerror.h"
}

// LEXER
COMMENT     : '//' ~('\n'|'\r')* '\r'? '\n' {$channel=HIDDEN;};

//All objects
OBJECTS     : 'objects';
NODES       : 'nodes';
EDGES       : 'edges';
    
//Node and edge attributes
LABEL       : 'label';
COLOR       : 'color' | 'colour';
NEIGHBOURS  : 'neighbours'|'neighbors';
SOURCE      : 'source';
TARGET      : 'target';
INCIDENT_EDGES    : 'incident edges';

//Number of nodes,columns,etc.
NUMOF_NODES : NUMOF NODES;
NUMOF_EDGES : NUMOF EDGES;
NUMOF_ROWS  : NUMOF ROWS;
NUMOF_COLS  : NUMOF COLUMNS;

 //Colors
fragment RED         : 'red';
fragment GREEN       : 'green';
fragment BLUE        : 'blue';
fragment YELLOW      : 'yellow';
fragment BLACK       : 'black';
fragment WHITE       : 'white';
fragment TEAL        : 'teal';
fragment PURPLE      : 'purple';
fragment GRAY        : 'gray';

fragment DIGIT       : '0'..'9';
fragment LETTER      : ('a'..'z'|'A'..'Z');

INT         : (DIGIT)+;
COLORNAME   : (RED|GREEN|BLUE|YELLOW|BLACK|WHITE|TEAL|PURPLE|GRAY);

PARENS_OPEN : '(';
PARENS_CLOSE: ')';

BRACK_OPEN  : '[';
BRACK_CLOSE : ']';

PLUS    : '+';
DPLUS   : '++';
MINUS   : '-';
TIMES   : '*';
DIV     : '/';
DOT     : '.';

MAXFUNC     : 'max';
MINFUNC     : 'min';
INTEGER : 'int';

NAME        : LETTER (LETTER|'_')* INT?;

WS          : ('\t' | ' ' | '\r')+ {$channel=HIDDEN;};


//PARSER
stmt            : (new_var LINEBREAK) -> ^(new_var) |  (expr LINEBREAK) -> ^(expr) | LINEBREAK;
expr            : assign | loop;

//variables
new_var         : DEFVAR^ named_var;
named_var       : NAME^ var_def_for;
var             : NAME^ (var_rest_for)?;

//loops
loop   : FOR^ NAME IN loop_var loopbody END;

loop_var @init{int objectRest = 0;} : (loop_var_name | object_rest {objectRest = 1;})
                                            -> {objectRest == 1}? ^(FOO PARENS_OPEN object_rest PARENS_CLOSE)
                                            -> ^(loop_var_name);

loop_var_name    : (var|int_rest|COLUMNS|ROWS);

loopbody        : DO^ (LINEBREAK!+) (expr(LINEBREAK!+))+;

sumloop         : SUM^ var;

//sumloop @init {int bodyExists = 0;}
//                : SUM NAME IN var (body=sumloopbody {bodyExists=1;} END)?
//                    -> {bodyExists == 1}? ^(SUM NAME IN var $body END)
//                    -> ^(SUM NAME IN var ^(WITH NAME) END);

//sumloopbody     : WITH^ (LINEBREAK!+) (mathexpr (LINEBREAK!+));

//assign
assign          : mathexpr (EQUALS|LEQ|GEQ|LESSER|GREATER)^ mathexpr;

//var restrictions
var_def_for     : PARENS_OPEN (row_rest | col_rest | int_rest | object_rest) (COMMA (row_rest | col_rest | int_rest | object_rest))* PARENS_CLOSE;
var_rest_for    : PARENS_OPEN (NAME | COLUMNS | ROWS | mathexpr | int_rest | object_rest | BLANK)
                                (COMMA (NAME | COLUMNS | ROWS | mathexpr | int_rest | object_rest | BLANK))* PARENS_CLOSE;

//restrictions for rows and columns
row_rest        : ROWS^ int_rest?;
col_rest        : COLUMNS^ int_rest?;

//restrictions for integer sets
int_rest        : closed_intvl -> ^(INTVL closed_intvl);
closed_intvl    : BRACK_OPEN mathexpr ( DOTS mathexpr )? BRACK_CLOSE;

object_rest     : with_rest -> ^(OBJREST with_rest);
with_rest       : graph_op_rest (with_or_without^)*;
graph_op_rest   : (of^)* object;
object          : NAME|NODES|EDGES|OBJECTS|PARENS_OPEN^ with_rest PARENS_CLOSE ;

with_or_without : (WITH|WITHOUT)^ attribute_spec;
of              : graph_op OF -> ^(OF graph_op);

//restrictions for object sets
//object_rest @init{int graphopExists = 0; int attributeSpecExists = 0;}
//                : (graph_op OF {graphopExists=1;})? object=(NAME|NODES|EDGES|OBJECTS) (attributeop=(WITH|WITHOUT) attribute_spec {attributeSpecExists = 1;})?
//                    -> {graphopExists == 1 && attributeSpecExists == 1}? ^(OBJREST ^($object ^(OF graph_op) ^($attributeop attribute_spec)))
//                    -> {graphopExists == 1}? ^(OBJREST ^($object ^(OF graph_op)))
//                    -> {attributeSpecExists == 1}? ^(OBJREST ^($object ^($attributeop attribute_spec)))
//                    -> ^(OBJREST $object);

attribute_spec  : label_spec | color_spec | node_spec | edge_spec;
graph_op        : (INCIDENT_EDGES|NEIGHBOURS|SOURCE|TARGET)^;

//attribute specifier
label_spec      : LABEL^ NAME;
color_spec      : COLOR^ COLORNAME;
node_spec       : NODES^ NAME;
edge_spec       : EDGES^ NAME;

//math expressions
mathexpr        : add -> ^(MATH add);
add             : mult ((PLUS|MINUS|DPLUS)^ mult)*;
mult            : atom ((TIMES|DIV)^ atom)*;
atom            : number | var | PARENS_OPEN^ add PARENS_CLOSE | NUMOF_COLS | NUMOF_ROWS | NUMOF_NODES | NUMOF_EDGES | mathfunc | sumloop;
mathfunc        : op=(MAXFUNC|MINFUNC|INTEGER) PARENS_OPEN add (COMMA add)* PARENS_CLOSE -> ^(MATHFUNC ^($op add*));
number          : INT^ (DOT INT)?;

//needed to generate the ast tree
ilp : (stmt {pANTLR3_STRING s = $stmt.tree->toStringTree($stmt.tree);
             printf(" tree \%s\n", s->chars);
            }
        )+;
