#include "gdslprogram.hpp"

#include "../parseexception.hpp"

using namespace gdsl;

const std::string GDSLProgram::GDSLNODES    = "_nodes";
const std::string GDSLProgram::GDSLEDGES    = "_edges";
const std::string GDSLProgram::GDSLCOLUMNS  = "_columns";
const std::string GDSLProgram::GDSLROWS     = "_rows";
const std::string GDSLProgram::GDSLBIGGESTINTVL = "_intvl";
const std::string GDSLProgram::GDSLOBJECTS = "_objects";
const std::vector<std::string> GDSLProgram::DEFAULT_VARS({GDSLProgram::GDSLNODES,
                                                          GDSLProgram::GDSLEDGES,
                                                          GDSLProgram::GDSLOBJECTS,
                                                          GDSLProgram::GDSLCOLUMNS,
                                                          GDSLProgram::GDSLROWS,
                                                          GDSLProgram::GDSLBIGGESTINTVL
                                                         });

bool GDSLProgram::isDefaultVar(std::shared_ptr<GDSLVariable> var)
{
    if(var->id() != GDSLProgram::GDSLBIGGESTINTVL && var->id() != GDSLProgram::GDSLCOLUMNS && var->id() != GDSLProgram::GDSLROWS
        && var->id() != GDSLProgram::GDSLNODES && var->id() != GDSLProgram::GDSLEDGES && var->id() != GDSLProgram::GDSLOBJECTS)
        return false;

    return true;
}

GDSLProgram::GDSLProgram(string &name, string &program)
{
    m_programStr = program;
    inputStr = (pANTLR3_UINT8)m_programStr.c_str();

    m_name = name;
    nameStr = (pANTLR3_UINT8)m_name.c_str();

    input = antlr3StringStreamNew(inputStr, ANTLR3_ENC_8BIT, strlen((char*)inputStr), nameStr);
    lex = GDSLGrammarLexerNew(input);
    tokens = antlr3CommonTokenStreamSourceNew(ANTLR3_SIZE_HINT, TOKENSOURCE(lex));
    parser = GDSLGrammarParserNew(tokens);

    m_r = parser->ilp(parser);
    m_tree = m_r.tree;

    parseFromTree(m_tree, m_variables);
}

GDSLProgram::~GDSLProgram()
{
    if(parser)
        parser->free(parser);
    if(tokens)
        tokens->free(tokens);
    if(lex)
        lex->free(lex);
    if(input)
        input->free(input);
}

bool GDSLProgram::parseFromTree(pANTLR3_BASE_TREE tree, GDSLVariable::GDSLVarTable& vars)
{
    if(vars.find(GDSLProgram::GDSLNODES) == vars.end())
    {
        GDSLRestList nodeRest;
        nodeRest.push_back(makeNodeNullRestriction(tree, vars));

        vars[GDSLProgram::GDSLNODES] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLNODES, nodeRest);
    }

    if(vars.find(GDSLProgram::GDSLEDGES) == vars.end())
    {
        GDSLRestList edgeRest;
        edgeRest.push_back(makeEdgeNullRestriction(tree, vars));

        vars[GDSLProgram::GDSLEDGES] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLEDGES, edgeRest);
    }

    if(vars.find(GDSLProgram::GDSLOBJECTS) == vars.end())
    {
        GDSLRestList objRest;
        objRest.push_back(makeObjectNullRestriction(tree, vars));

        vars[GDSLProgram::GDSLOBJECTS] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLOBJECTS, objRest);
    }

    if(vars.find(GDSLProgram::GDSLCOLUMNS) == vars.end())
    {
        GDSLRestList columnRest;
        columnRest.push_back(makeMaxColumnRestriction(tree, vars));

        vars[GDSLProgram::GDSLCOLUMNS] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLCOLUMNS, columnRest);
    }

    if(vars.find(GDSLProgram::GDSLROWS) == vars.end())
    {
        GDSLRestList rowRest;
        rowRest.push_back(makeMaxRowRestriction(tree, vars));

        vars[GDSLProgram::GDSLROWS] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLROWS, rowRest);
    }

    if(vars.find(GDSLProgram::GDSLBIGGESTINTVL) == vars.end())
    {
        GDSLRestList intvlRest;
        intvlRest.push_back(makeMaxIntRestriction(tree, vars));

        vars[GDSLProgram::GDSLBIGGESTINTVL] = std::make_shared<GDSLVariable>(GDSLProgram::GDSLBIGGESTINTVL, intvlRest);
    }

    int currentLine = -1;
    std::function<bool (pANTLR3_BASE_TREE)> parse = [&] (pANTLR3_BASE_TREE tree)
    {
        currentLine = tree->getLine(tree) == 0 ? ++currentLine : tree->getLine(tree);
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        if(tok)
        {
            switch(tok->type)
            {
                case DEFVAR:
                {
                    auto variable = std::make_shared<GDSLVariable>(tree, nullptr, vars);
                    vars[variable->id()] = variable;

                    variable.reset();

                    return true;
                }
                case EQUALS:
                {
                    if(createAssign<Equal>(tree, vars))
                        return true;

                    return false;
                }
                case LEQ:
                {
                    if(createAssign<Leq>(tree, vars))
                        return true;

                    return false;
                }
                case GEQ:
                {
                    if(createAssign<Geq>(tree, vars))
                        return true;

                    return false;
                }
                case LESSER:
                {
                    if(createAssign<Lesser>(tree, vars))
                        return true;

                    return false;
                }
                case GREATER:
                {
                    if(createAssign<Greater>(tree, vars))
                        return true;

                    return false;
                }
                case FOR:
                {
                    auto loop = std::make_shared<GDSLLoop>(tree, vars);
                    m_program.push_back(loop);

                    return true;
                }
                case LINEBREAK:
                {
                    ++currentLine;
                    return true;
                }
                default:
                {
                    if(tree->getLine(tree) == 0)
                        ++currentLine;

                    throw ParseException("This line is utter crap!", currentLine);
                }
            }
        }
        else
        {
            if(antlrutils::foreachChildDo(tree, parse))
                return true;

            if(tree->getLine(tree) == 0)
                ++currentLine;

            throw ParseException("This line is utter crap!", currentLine);
        }
    };

    antlrutils::foreachChildDo(tree, parse);

    return true;
}

void GDSLProgram::eval()
{
    eval(m_variables, m_constants, m_constraints, m_attr, m_grid);
}

void GDSLProgram::runForGraph(std::shared_ptr<RandomAccessGraph> attr, std::shared_ptr<Grid> grid)
 {
    m_attr = attr;
    m_grid = grid;

    clear();

    for(auto defVar : DEFAULT_VARS)
        m_variables[defVar]->apply(m_variables, m_constants, m_attr, m_grid);

    for(auto var : m_variables)
        if(std::find(DEFAULT_VARS.begin(), DEFAULT_VARS.end(), var.first) == DEFAULT_VARS.end())
            var.second->apply(m_variables, m_constants, m_attr, m_grid);

    for(auto var : m_variables)
        var.second->createIlpVariables();

    eval();
}

void GDSLProgram::eval(GDSLVariable::GDSLVarTable &scope, ConstantsMap &constants, EquationList &constraints, std::shared_ptr<RandomAccessGraph> attr, std::shared_ptr<Grid> grid)
{
    for(auto expr : m_program)
        expr->eval(scope, constants, constraints, attr, grid);
}

std::string GDSLProgram::toString() const
{
    std::stringstream ss;

    for(auto var : m_variables)
        ss << var.first << " : " << var.second.get()->toString() << std::endl;

    for(auto expr : m_program)
        ss << expr->toString() << std::endl;

    return ss.str();
}

std::shared_ptr<GRBModel> GDSLProgram::toGurobi(GRBEnv& env) const
{
    std::shared_ptr<GRBModel> model = std::make_shared<GRBModel>(env);

    for(auto var : m_variables)
    {
        if(!GDSLProgram::isDefaultVar(var.second))
            var.second->createGRBVars(model);
    }

    model->update();

    for(auto eq : m_constraints)
        eq->toGurobi(model);

    model->update();

    return model;
}

string GDSLProgram::grbToString() const
{
    std::stringstream ss;

    for(auto var : m_variables)
        if(!GDSLProgram::isDefaultVar(var.second))
            ss << var.second->toString() << std::endl << var.second->grbToString() << std::endl;

    return ss.str();
}

const std::vector<std::shared_ptr<GDSLVariable>> GDSLProgram::variables()
{
    std::vector<std::shared_ptr<GDSLVariable>> vars;
    for(auto var : m_variables)
        vars.push_back(var.second);

    return vars;
}

void GDSLProgram::clear()
{
    for(auto var : m_variables)
        var.second->clearIlpVars();

    for(auto expr : m_program)
        expr->clear();

    m_constraints.clear();
}

template<class OperatorType>
bool GDSLProgram::createAssign(pANTLR3_BASE_TREE tree, GDSLVariable::GDSLVarTable& vars)
{
    std::shared_ptr<Operator> op =  std::make_shared<OperatorType>();
    auto assign = std::make_shared<GDSLAssign>(tree, vars, op);
    m_program.push_back(assign);

    return true;
}
