#include "mathexpr.hpp"

#include "../parseexception.hpp"

using namespace gdsl;

bool MathExpr::parseFromTree(pANTLR3_BASE_TREE tree, GDSLVariable::GDSLVarTable&)
{
    m_tree = tree;
    m_evaled = false;

    return true;
}

//not really useful, perhaps set everything to zero
void MathExpr::eval(GDSLVariable::GDSLVarTable &scope, ConstantsMap &constants, GDSLExpr::EquationList&, std::shared_ptr<RandomAccessGraph> attr,
                    std::shared_ptr<Grid> grid)
{
    auto exprs = eval(scope, constants, attr, grid);
    return;
}

std::shared_ptr<IlpExprList> MathExpr::eval(GDSLVariable::GDSLVarTable &scope, ConstantsMap &constants, std::shared_ptr<RandomAccessGraph> attr, std::shared_ptr<Grid> grid)
{
    std::function<std::shared_ptr<IlpExprList>(pANTLR3_BASE_TREE)> parse;
    std::function<std::shared_ptr<IlpExprList>(pANTLR3_BASE_TREE)> parseMathfunc;
    std::function<bool (pANTLR3_BASE_TREE)> parseArgument;

    bool interpretNamesAsInteger = false;

    parse = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
        if(tok)
        {
            switch(tok->type)
            {
                case PLUS:
                {
                    pANTLR3_BASE_TREE left = antlrutils::getChild(tree, 0);
                    auto leftValue = parse(left);

                    pANTLR3_BASE_TREE right = antlrutils::getChild(tree, 1);
                    auto rightValue = parse(right);

                    return leftValue + rightValue;
                }
                case DPLUS:
                {
                    pANTLR3_BASE_TREE left = antlrutils::getChild(tree, 0);
                    auto leftValue = parse(left);

                    pANTLR3_BASE_TREE right = antlrutils::getChild(tree, 1);
                    auto rightValue = parse(right);

                    return leftValue % rightValue;
                }
                case MINUS:
                {
                    pANTLR3_BASE_TREE left = antlrutils::getChild(tree, 0);
                    auto leftValue = parse(left);

                    pANTLR3_BASE_TREE right = antlrutils::getChild(tree, 1);
                    auto rightValue = parse(right);

                    auto minusRight = -1 * rightValue;

                    return leftValue + minusRight;
                }
                case TIMES:
                {
                    pANTLR3_BASE_TREE left = antlrutils::getChild(tree, 0);
                    auto leftValue = parse(left);

                    pANTLR3_BASE_TREE right = antlrutils::getChild(tree, 1);
                    auto rightValue = parse(right);

                    return leftValue * rightValue;
                }
                case DIV:
                {
                    pANTLR3_BASE_TREE left = antlrutils::getChild(tree, 0);
                    auto leftValue = parse(left);

                    pANTLR3_BASE_TREE right = antlrutils::getChild(tree, 1);
                    auto rightValue = parse(right);

                    return leftValue / rightValue;
                }
                case NUMOF_ROWS:
                {
                    auto expr = std::make_shared<IlpExpr>(grid->y());
                    auto exprs = std::make_shared<IlpExprList>();
                    exprs->append(expr);

                    return exprs;
                }
                case NUMOF_COLS:
                {
                    auto expr = std::make_shared<IlpExpr>(grid->x());
                    auto exprs = std::make_shared<IlpExprList>();
                    exprs->append(expr);

                    return exprs;
                }
                case NUMOF_NODES:
                {
                    auto expr = std::make_shared<IlpExpr>(attr->attr()->constGraph().numberOfNodes());
                    auto exprs = std::make_shared<IlpExprList>();
                    exprs->append(expr);

                    return exprs;
                }
                case NUMOF_EDGES:
                {
                    auto expr = std::make_shared<IlpExpr>(attr->attr()->constGraph().numberOfEdges());
                    auto exprs = std::make_shared<IlpExprList>();
                    exprs->append(expr);

                    return exprs;
                }
                case MATHFUNC:
                {
                    pANTLR3_BASE_TREE func = antlrutils::getChild(tree, 0);
                    auto res =  parseMathfunc(func);
                    return res;
                }
                case INT:
                {
                    std::string decimalPlace = "";
                    if(tree->getChildCount(tree) == 2)
                        decimalPlace = antlrutils::getText(antlrutils::getChild(tree, 1));

                    std::stringstream ss;
                    double num = 0;

                    ss << antlrutils::getText(tree) << '.' << decimalPlace;
                    ss >> num;

                    auto expr = std::make_shared<IlpExpr>(num);
                    auto exprs = std::make_shared<IlpExprList>();
                    exprs->append(expr);

                    return exprs;
                }
                case MATH:
                {
                    return parse(antlrutils::getChild(tree, 0));
                }
                case PARENS_OPEN:
                {
                    return parse(antlrutils::getChild(tree, 0));
                }
                case SUM:
                {
                    std::shared_ptr<SumExpr> sum = std::make_shared<SumExpr>(tree, scope);
                    return sum->eval(scope, constants, attr, grid);
                }
                case NAME:
                {
                    std::string name = antlrutils::getText(tree);
                    auto exprs = std::make_shared<IlpExprList>();

                    if(constants.find(name) != constants.end())
                    {
                        int num = constants[name];

                        auto expr = std::make_shared<IlpExpr>(num);
                        exprs->append(expr);

                        return exprs;
                    }

                    if(scope.find(name) == scope.end())
                        throw ParseException("Unknown Variable " + name, m_tree->getLine(m_tree));

                    auto scopeVar = scope.find(name)->second;
                    auto var = std::make_shared<GDSLVariable>(tree, scopeVar, scope);
                    var->apply(scope, constants, attr, grid);
                    auto rests = var->restrictions();

                    auto ilpVars = scopeVar->get(rests);

                    if(interpretNamesAsInteger)
                    {
                        if(rests.size() != 1)
                            throw ParseException(antlrutils::getText(tree) + " because too many restrictions were given to convert into number.", tree->getLine(tree));

                        try
                        {
                            int num = rests[0]->asInteger();
                            auto expr = std::make_shared<IlpExpr>(num);
                            exprs->append(expr);
                        }
                        catch(NotAnIntegerException e)
                        {
                            throw ParseException(antlrutils::getText(tree) + " could not be interpreted as integer.", tree->getLine(tree));
                        }
                    }
                    else
                    {
                        for(auto ilpVar : ilpVars)
                        {
                            auto expr = std::make_shared<IlpExpr>(0);
                            expr->add(1, ilpVar);
                            exprs->append(expr);
                        }
                    }

                    return exprs;
                }
                default:
                {
                    throw ParseException(antlrutils::getText(tree) + " while evaluating math expression.", tree->getLine(tree));
                }
            }
        }
    };

    parseMathfunc = [&] (pANTLR3_BASE_TREE tree)
    {
        std::function<std::shared_ptr<IlpExprList>(std::vector<std::shared_ptr<IlpExprList>>&)> mathfunc;
        std::vector<std::shared_ptr<IlpExprList>> args;
        std::shared_ptr<IlpExprList> result;

        parseArgument = [&] (pANTLR3_BASE_TREE tree)
        {
            pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
            if(tok)
            {
                interpretNamesAsInteger = true;
                std::shared_ptr<IlpExprList> arg = parse(tree);
                interpretNamesAsInteger = false;

                if(!arg->isNumber())
                    throw ParseException(antlrutils::getText(tree) + arg->toString() + " is not a number", tree->getLine(tree));

                args.push_back(arg);
            }

            return true;
        };

        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
        if(tok)
        {
            switch(tok->type)
            {
                case MAXFUNC:
                {
                    mathfunc = [&] (std::vector<std::shared_ptr<IlpExprList>>& args)
                    {
                        std::shared_ptr<IlpExprList> max = nullptr;

                        for(std::shared_ptr<IlpExprList> arg : args)
                        {
                            if(max == nullptr)
                                max = arg;

                            else if(arg > max)
                                max = arg;
                        }

                        return max;
                    };

                    break;
                }
                case MINFUNC:
                {
                    mathfunc = [&] (std::vector<std::shared_ptr<IlpExprList>>& args)
                    {
                        std::shared_ptr<IlpExprList> min = nullptr;

                        for(std::shared_ptr<IlpExprList> arg : args)
                        {
                            if(min == nullptr)
                                min = arg;

                            else if(arg < min)
                                min = arg;
                        }

                        return min;
                    };

                    break;
                }
                case INTEGER:
                {
                    mathfunc = [&] (std::vector<std::shared_ptr<IlpExprList>>& args)
                    {
                        if(args.size() != 1)
                            throw ParseException(antlrutils::getText(tree) + " while evaluating math function. Too many arguments for int().", tree->getLine(tree));

                        return args.at(0);
                    };

                    break;
                }
                default:
                {
                    throw ParseException(antlrutils::getText(tree) + " while evaluating math function.", tree->getLine(tree)); //TODO: Correct error message
                }
            }

            antlrutils::foreachChildDo(tree, parseArgument);
            result = mathfunc(args);

            return result;
        }

        return result;
    };

    std::shared_ptr<IlpExprList> exprs = parse(m_tree);
    return exprs;
}

std::shared_ptr<IlpExprList> gdsl::operator+(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    std::shared_ptr<IlpExprList> exprs = std::make_shared<IlpExprList>();

    auto leftIt = left->begin();
    auto rightIt = right->begin();

    if(left->isNumber())
        for(;rightIt < right->end(); ++rightIt)
            exprs->append(*leftIt + *rightIt);

    else if(right->isNumber())
        for(;leftIt < left->end(); ++leftIt)
            exprs->append(*leftIt + *rightIt);

    else
        for(; leftIt < left->end() && rightIt < right->end(); ++leftIt, ++rightIt)
            exprs->append(*leftIt + *rightIt);

    return exprs;
}

std::shared_ptr<IlpExprList> gdsl::operator%(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    std::shared_ptr<IlpExprList> exprs = std::make_shared<IlpExprList>();

    for(auto leftIt = left->begin(); leftIt < left->end() ; ++leftIt)
        for(auto rightIt = right->begin(); rightIt < right->end(); ++rightIt)
            exprs->append(*leftIt + *rightIt);

    return exprs;
}

std::shared_ptr<IlpExprList> gdsl::operator-(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    std::shared_ptr<IlpExprList> exprs = std::make_shared<IlpExprList>();

    auto leftIt = left->begin();
    auto rightIt = right->begin();

    if(left->isNumber())
        for(;rightIt < right->end(); ++rightIt)
            exprs->append(*leftIt - *rightIt);

    else if(right->isNumber())
        for(;leftIt < left->end(); ++leftIt)
            exprs->append(*leftIt - *rightIt);

    else
        for(; leftIt < left->end() && rightIt < right->end(); ++leftIt, ++rightIt)
            exprs->append(*leftIt - *rightIt);

    return exprs;
}

std::shared_ptr<IlpExprList> gdsl::operator*(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    if(left->isNumber())
        return left->at(0)->getConstant() * right;

    else if(right->isNumber())
        return right->at(0)->getConstant() * left;

    else
        throw ParseException("Invalid multiplication of two variables!", 0);
}

std::shared_ptr<IlpExprList> gdsl::operator*(const std::shared_ptr<IlpExprList> &right, double mult)
{
    return mult * right;
}

std::shared_ptr<IlpExprList> gdsl::operator*(double mult, const std::shared_ptr<IlpExprList> &right)
{
    std::shared_ptr<IlpExprList> exprs = std::make_shared<IlpExprList>();

    auto rightIt = right->begin();

    for(; rightIt < right->end(); ++rightIt)
        exprs->append(mult * *rightIt);

    return exprs;
}

std::shared_ptr<IlpExprList> gdsl::operator/(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    std::shared_ptr<IlpExprList> exprs = std::make_shared<IlpExprList>();

    auto leftIt = left->begin();
    auto rightIt = right->begin();

    for(; leftIt < left->end() && rightIt < right->end(); ++leftIt, ++rightIt)
        exprs->append(*leftIt / *rightIt);

    return exprs;
}


bool gdsl::operator<(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    auto leftIt = left->begin();
    auto rightIt = right->begin();

    return *leftIt < *rightIt;
}


bool gdsl::operator>(const std::shared_ptr<IlpExprList> &left, const std::shared_ptr<IlpExprList> &right)
{
    auto leftIt = left->begin();
    auto rightIt = right->begin();

    return *leftIt > *rightIt;
}
