#include "gdslvariable.hpp"

#include <iostream>
#include "expressions/gdslprogram.hpp"
#include "parseexception.hpp"

using namespace gdsl;

GDSLVariable::GDSLVariable(const string &name, GDSLRestList &rests) :
    m_name(name),
    m_rest(rests)
{
}

GDSLVariable::GDSLVariable(pANTLR3_BASE_TREE tree, std::shared_ptr<GDSLVariable> orig, GDSLVarTable &scope)
{
    parseFromTree(tree, orig, scope);
}

bool GDSLVariable::parseFromTree(pANTLR3_BASE_TREE tree, std::shared_ptr<GDSLVariable> orig, GDSLVarTable &scope)
{
    int currentLine = tree->getLine(tree);

    pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
    if(tok->type == NAME)
        m_name = antlrutils::getText(tree);

    std::function<std::shared_ptr<GDSLAttributeRest> (pANTLR3_BASE_TREE)> parseAttributeRest = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case COLOR:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building Colorrestriction", currentLine);

                return makeNodeColorRestriction(tree, scope);
            }
            case LABEL:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building labelrestriction", currentLine);

                return makeNodeLabelRestriction(tree, scope);
            }
            default:
            {
                throw ParseException(antlrutils::getText(tree) + " while building attribute restriction", currentLine);
            }
        }
    };

    std::function<std::shared_ptr<GDSLAttributeRest> (pANTLR3_BASE_TREE)> parseNegatedAttributeRest = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case COLOR:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building Colorrestriction", currentLine);

                return makeNegatedNodeColorRestriction(tree, scope);
            }
            case LABEL:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building labelrestriction", currentLine);

                return makeNegatedNodeLabelRestriction(tree, scope);
            }
            case NODES:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building labelrestriction", currentLine);

                return makeWithoutNodesRestriction(tree, scope);
            }
            case EDGES:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree) + " while building labelrestriction", currentLine);

                return makeWithoutEdgesRestriction(tree, scope);
            }
            default:
            {
                throw ParseException(antlrutils::getText(tree) + " while building attribute restriction", currentLine);
            }
        }
    };

    //TODO: Differentiate between node and edge restrictions, error handling
    std::function<std::shared_ptr<GDSLPropertyRestriction> (pANTLR3_BASE_TREE)> parseGraphOp = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case NEIGHBOURS:
            {
                return makeNeighbourRestriction();
            }
            case SOURCE:
            {
                return makeSourceRestriction();
            }
            case TARGET:
            {
                return makeTargetRestriction();
            }
            case INCIDENT_EDGES:
            {
                return makeIncidentEdgesRestriction();
            }
            default:
            {
                throw ParseException(antlrutils::getText(tree) + " while building graphop restriction", currentLine);
            }
        }
    };

    std::function<std::shared_ptr<GDSLRestriction> (pANTLR3_BASE_TREE)> parseObjectRest = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
        std::shared_ptr<GDSLRestriction> op;

        switch(tok->type)
        {
            case NAME:
            {
                op = makeNameNullRestriction(tree, scope);
                break;
            }
            case NODES:
            {
                op = makeNodeNullRestriction(tree, scope);
                break;
            }
            case EDGES:
            {
                op = makeEdgeNullRestriction(tree, scope);
                break;
            }
            case OBJECTS:
            {
                op = makeObjectNullRestriction(tree, scope);
                break;
            }
            case OF:
            {
                if(tree->getChildCount(tree) != 2)
                    throw ParseException("Expected at least two arguments for restriction! " + antlrutils::getText(tree), currentLine);

                auto inner = parseObjectRest(antlrutils::getChild(tree, 1));
                op = parseGraphOp(antlrutils::getChild(tree, 0));
                op->setInnerRestriction(inner);
                break;
            }
            case WITH:
            {
                if(tree->getChildCount(tree) != 2)
                    throw ParseException("Expected at least two arguments for restriction! " + antlrutils::getText(tree), currentLine);

                auto inner = parseObjectRest(antlrutils::getChild(tree, 1));
                op = parseAttributeRest(antlrutils::getChild(tree, 0));
                op->setInnerRestriction(inner);
                break;
            }
            case WITHOUT:
            {
                if(tree->getChildCount(tree) != 2)
                    throw ParseException("Expected at least two arguments for restriction! " + antlrutils::getText(tree), currentLine);

                auto inner = parseObjectRest(antlrutils::getChild(tree, 1));
                op = parseNegatedAttributeRest(antlrutils::getChild(tree, 0));
                op->setInnerRestriction(inner);
                break;
            }
            case PARENS_OPEN:
            {
                if(tree->getChildCount(tree) != 2)
                    throw ParseException("Expected at least two arguments for restriction! " + antlrutils::getText(tree), currentLine);

                return parseObjectRest(antlrutils::getChild(tree, 0));
            }
            default:
            {
                throw ParseException(antlrutils::getText(tree) + " while building object restriction", currentLine);
            }
        }

        return op;
    };

    std::function<bool (pANTLR3_BASE_TREE)> parse = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case NAME:
            {
                std::string text = antlrutils::getText(tree);

                if(m_name == "")
                {
                    m_name = text;
                    antlrutils::foreachChildDo(tree, parse);
                }
                else if(scope.find(text) != scope.end())
                {
                    for(auto rest : scope[text]->restrictions())
                        m_rest.push_back(rest);
                }

                return true;
            }
            case COLUMNS:
            {
                if(tree->getChildCount(tree) > 0)
                    m_rest.push_back(makeColumnRestriction(antlrutils::getChild(tree, 0), scope));
                else
                    m_rest.push_back(makeMaxColumnRestriction(antlrutils::getChild(tree, 0), scope));

                return true;
            }
            case ROWS:
            {
                if(tree->getChildCount(tree) > 0)
                    m_rest.push_back(makeRowRestriction(antlrutils::getChild(tree, 0), scope));
                else
                    m_rest.push_back(makeMaxRowRestriction(antlrutils::getChild(tree, 0), scope));

                return true;
            }
            case INTVL:
            {
                if(orig != nullptr)
                {
                    auto origRest = orig->restrictions().at(m_rest.size());

                    if(origRest->outputType() == RestrictionType::COLUMNREST)
                        m_rest.push_back(makeColumnRestriction(tree, scope));
                    else if(origRest->outputType() == RestrictionType::ROWREST)
                        m_rest.push_back(makeRowRestriction(tree, scope));
                    else if(origRest->outputType() == RestrictionType::INTREST)
                        m_rest.push_back(makeIntRestriction(tree, scope));
                    else
                        return false;
                }
                else
                    m_rest.push_back(makeIntRestriction(tree, scope));

                return true;
            }
            case MATH:
            {
                m_rest.push_back(makeIntRestriction(tree, scope));

                return true;
            }
            case BLANK:
            {
                std::shared_ptr<GDSLRestriction> rest = orig->getRestrictionAt(m_rest.size());
                m_rest.push_back(rest);

                return true;
            }
            case OBJREST:
            {
                if(tree->getChildCount(tree) != 1)
                    throw ParseException(antlrutils::getText(tree), currentLine);

                auto objRest = parseObjectRest(antlrutils::getChild(tree, 0));

                if(m_name == "")
                {
                    if(objRest->outputType() == RestrictionType::NODEREST)
                        m_name = GDSLProgram::GDSLNODES;
                    else if(objRest->outputType() == RestrictionType::EDGEREST)
                        m_name = GDSLProgram::GDSLEDGES;
                    else if(objRest->outputType() == RestrictionType::OBJECTREST)
                        m_name = GDSLProgram::GDSLOBJECTS;
                    else
                        ; //Error
                }

                m_rest.push_back(objRest);

                return true;
            }
            default:
            {
                if(tok->type == PARENS_OPEN || tok->type == PARENS_CLOSE || tok->type == COMMA)
                    return true;

                throw ParseException(antlrutils::getText(tree)  + " while building restriction", currentLine);
            }
        }
    };

    if(tok->type == INTVL || tok->type == OBJREST)
        parse(tree);
    else
        antlrutils::foreachChildDo(tree, parse);

    if(orig != nullptr)
    {
        //No restrictions were given assume maximal restrictions are ment
        if(m_rest.size() == 0)
        {
            for(auto rest : orig->restrictions())
                m_rest.push_back(rest);
        }
        //If no match and not equal zero -> error
        else if(m_rest.size() != orig->restrictions().size())
        {
            std::stringstream ss;
            ss << "Number of restrictions after parsing variable " << m_name << " is " << m_rest.size()
                  << " should be " << orig->restrictions().size();
            throw ParseException(ss.str(), currentLine);
        }
    }

    return true;
}

std::string GDSLVariable::toString() const
{
    std::stringstream ss;
    ss << m_name << "(";

    for(auto it = m_rest.begin(); it < m_rest.end(); ++it)
    {
        ss << (*it)->toString();

        if(it + 1 < m_rest.end())
            ss << ",";
    }

    ss << ")";

    return ss.str();
}

std::string GDSLVariable::toString(int id) const
{
    std::stringstream ss;
    ss << m_name << "(";
    auto rests = restFromId(id);

    for(auto it = rests->begin(); it < rests->end(); ++it)
    {
        ss << (*it)->toString();

        if(it + 1 < rests->end())
            ss << ",";
    }

    ss << ")";

    return ss.str();
}

bool GDSLVariable::restListIsSubset(const GDSLRestList &rest) const
{
    if(rest.size() != m_rest.size())
        return false;

    bool isSubset = true;
    int i = 0;
    for(auto r : rest)
    {
        isSubset &= r.get()->isSubset(m_rest.at(i));
    }

    return isSubset;
}

GDSLVariable::IlpVarList GDSLVariable::get(GDSLRestList &rests)
{
    CombinationIterator combinations(rests);
    IlpVarList ilpVars;

    while(!combinations.isEnd())
    {
        GDSLRestList singleRests = combinations.current();
        int id = idFromRest(singleRests);
        ilpVars.push_back(m_ilpVars.at(id));

        combinations.next();
    }

    return ilpVars;
}

void GDSLVariable::createIlpVariables()
{
    int num = 1;
    for(auto rest : m_rest)
        num *= rest->length();

    for(int i = 0; i < num; ++i)
    {
        auto var = std::make_shared<IlpVariable>(shared_from_this(), i);
        m_ilpVars.push_back(var);
    }
}

void GDSLVariable::clearIlpVars()
{
    m_ilpVars.clear();
}

void GDSLVariable::createGRBVars(std::shared_ptr<GRBModel> &model)
{
    for(auto var : m_ilpVars)
        var->createGRBVar(model);
}

string GDSLVariable::grbToString() const
{
    std::stringstream ss;

    for(std::shared_ptr<IlpVariable> var : m_ilpVars)
        ss << var->toString() << " " << var->grbToString() << std::endl;

    return ss.str();
}

void GDSLVariable::apply(GDSLVarTable &scope, ConstantsMap& constants, std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
{
    m_attr = attr;
    m_grid = grid;

    for(auto rest : m_rest)
        rest->apply(scope, constants, attr, grid);
}

int GDSLVariable::idFromRest(GDSLRestList &rests) const
{
    int id = 0;

    std::vector<int> restSizes;
    for(auto rest : m_rest)
        restSizes.push_back(rest->length());

    std::vector<int> dimPositions;
    auto m_restIt = m_rest.begin();
    auto restIt = rests.begin();

    for(; restIt < rests.end() && m_restIt < m_rest.end(); ++restIt, ++m_restIt)
        dimPositions.push_back((*m_restIt)->pos(*restIt));

    int dimDone = 1;
    for(auto dimIt = dimPositions.begin(); dimIt < dimPositions.end(); ++dimIt)
    {
        int factor = 1;
        for(auto sizeIt = restSizes.begin() + dimDone; sizeIt < restSizes.end(); ++sizeIt)
            factor *= (*sizeIt);

        id += factor * (*dimIt);
        ++dimDone;
    }

    return id;
}

std::shared_ptr<GDSLRestList> GDSLVariable::restFromId(int id) const
{
    std::shared_ptr<GDSLRestList> rests = std::make_shared<GDSLRestList>();

    std::vector<int> restSizes;
    for(auto rest : m_rest)
        restSizes.push_back(rest->length());

    std::vector<int> dimPositions;
    auto restIt = m_rest.begin();
    auto dimIt = dimPositions.begin();


    int dimDone = 1;
    for(restIt = m_rest.begin(); restIt < m_rest.end(); ++restIt)
    {
        int factor = 1;
        for(auto sizeIt = restSizes.begin() + dimDone; sizeIt < restSizes.end(); ++sizeIt)
            factor *= (*sizeIt);

        dimPositions.push_back((id/factor) % (*restIt)->length());
        ++dimDone;
    }

    for(dimIt = dimPositions.begin(), restIt = m_rest.begin(); dimIt < dimPositions.end() && restIt < m_rest.end(); ++dimIt, ++restIt)
    {
        auto pos = (*restIt)->begin() + (*dimIt);
        rests->push_back((*restIt)->singleRestAt(pos));
    }

    return rests;
}
