#include "ilpexpr.hpp"

using namespace gdsl;

IlpExpr::IlpExpr(std::vector<double> coeffs, std::vector< std::shared_ptr<IlpVariable> > variables, int constant)
{
    m_constant = constant;

    auto coeff = coeffs.begin();
    auto var = variables.begin();

    while(coeff < coeffs.end() && var < variables.end())
    {
        auto pair = std::make_pair(*coeff, *var);
        m_vars.push_back(pair);
        ++coeff;
        ++var;
    }
}

void IlpExpr::add(double coeff, std::shared_ptr<IlpVariable>& var)
{
    auto pair = std::make_pair(coeff, var);
    add(pair);
}

std::string IlpExpr::toString() const
{
    std::stringstream ss;

    for(auto pair : m_vars)
        ss << pair.first << "*" << pair.second->toString() << " + ";

    ss << m_constant;

    return ss.str();
}

GRBLinExpr IlpExpr::toGurobi() const
{
    GRBLinExpr expr;

    expr += m_constant;
    for(IlpVarCoeffPair pair : m_vars)
    {
        expr += pair.first * (*pair.second->toGurobi());
    }

    return expr;
}

std::shared_ptr<IlpExpr> gdsl::operator+(const std::shared_ptr<IlpExpr> &left, const std::shared_ptr<IlpExpr> &right)
{
    std::shared_ptr<IlpExpr> expr = std::make_shared<IlpExpr>();

    expr->setConstant(left->getConstant() + right->getConstant());

    for(auto leftExpr : *left)
        expr->add(leftExpr);

    for(auto rightExpr : *right)
        expr->add(rightExpr);

    return expr;
}

std::shared_ptr<IlpExpr> gdsl::operator-(const std::shared_ptr<IlpExpr> &left, const std::shared_ptr<IlpExpr> &right)
{
    std::shared_ptr<IlpExpr> expr = std::make_shared<IlpExpr>();

    expr->setConstant(left->getConstant() - right->getConstant());
    std::shared_ptr<IlpExpr> minRight = -1 * right;

    for(auto leftExpr : *left)
        expr->add(leftExpr);

    for(auto rightExpr : *minRight)
        expr->add(rightExpr);

    return expr;
}

std::shared_ptr<IlpExpr> gdsl::operator*(const std::shared_ptr<IlpExpr>& left, const double right)
{
    return right * left;
}

std::shared_ptr<IlpExpr> gdsl::operator*(const double left, const std::shared_ptr<IlpExpr> &right)
{
    std::shared_ptr<IlpExpr> expr = std::make_shared<IlpExpr>();

    expr->setConstant(left * right->getConstant());
    for(auto rightExpr : *right)
    {
        double coeff = rightExpr.first * left;
        auto pair = std::make_pair(coeff, rightExpr.second);
        expr->add(pair);
    }

    return expr;
}


std::shared_ptr<IlpExpr> gdsl::operator/(const std::shared_ptr<IlpExpr> &left, const std::shared_ptr<IlpExpr> &right)
{
    if(!right->isNumber())
        return left;

    return left * (1/right->getConstant());
}


bool gdsl::operator>(const std::shared_ptr<IlpExpr> &left, const std::shared_ptr<IlpExpr> &right)
{
    return left->getConstant() > right->getConstant();
}


bool gdsl::operator<(const std::shared_ptr<IlpExpr> &left, const std::shared_ptr<IlpExpr> &right)
{
    return left->getConstant() < right->getConstant();
}
