#include "gdslattributerest.hpp"

#include "../gdslvariable.hpp"

using namespace gdsl;

bool GDSLAttributeRest::parseFromTree(pANTLR3_BASE_TREE tree, std::unordered_map<string, std::shared_ptr<GDSLVariable> >&)
{
    std::function<bool (pANTLR3_BASE_TREE)> parse = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case COLORNAME:
            {
                m_attribute = antlrutils::getText(tree);

                return true;
            }
            case NAME:
            {
                m_attribute = antlrutils::getText(tree);

                return true;
            }
            default:
            {
               return true;
            }
        }

        return true;
    };

    if(tree->getChildCount(tree) > 0)
        return antlrutils::foreachChildDo(tree, parse);
    else
        return parse(tree);
}

std::shared_ptr<GDSLAttributeRest> gdsl::makeNodeColorRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>& attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        NodeFunction func = [&] (ogdf::node node)
        {
            ogdf::String nodeColor = attr->attr()->colorNode(node);
            int id = attr->idByNode(node);
            if(rest->colorName(nodeColor.cstr()) == rest->attribute())
                rest->append(id);
        };

        rest->forEachNodeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::NODEREST, RestrictionType::NODEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeEdgeColorRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>& attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            ogdf::String edgeColor = attr->attr()->colorEdge(edge);
            int id = attr->idByEdge(edge);
            if(rest->colorName(edgeColor.cstr()) == rest->attribute())
                rest->append(id);
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}

std::shared_ptr<GDSLAttributeRest> gdsl::makeNodeNullRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>& attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        NodeFunction func = [&] (ogdf::node node)
        {
            rest->append(attr->idByNode(node));
        };

        rest->forEachNodeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::NODEREST, RestrictionType::NODEREST, tree, scope);
}

std::shared_ptr<GDSLAttributeRest> gdsl::makeEdgeNullRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            rest->append(attr->idByEdge(edge));
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeObjectNullRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        std::function<void (ogdf::node)> funcNodes = [&] (ogdf::node node)
        {
            rest->append(attr->idByNode(node));
        };

        std::function<void (ogdf::edge)> funcEdges = [&] (ogdf::edge edge)
        {
            rest->append(attr->idByEdge(edge));
        };

        rest->forEachNodeDo(attr, funcNodes);
        rest->setNumNodes(rest->length());
        rest->forEachEdgeDo(attr, funcEdges);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::OBJECTREST, RestrictionType::OBJECTREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNegatedNodeColorRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        NodeFunction func = [&] (ogdf::node node)
        {
            ogdf::String nodeColor = attr->attr()->colorNode(node);
            int id = attr->idByNode(node);
            if(rest->colorName(nodeColor.cstr()) != rest->attribute())
                rest->append(id);
        };

        rest->forEachNodeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::NODEREST, RestrictionType::NODEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNegatedEdgeColorRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            ogdf::String edgeColor = attr->attr()->colorEdge(edge);
            int id = attr->idByEdge(edge);
            if(rest->colorName(edgeColor.cstr()) != rest->attribute())
                rest->append(id);
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNodeLabelRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        NodeFunction func = [&] (ogdf::node node)
        {
            ogdf::String nodeLabel = attr->attr()->labelNode(node);
            int id = attr->idByNode(node);
            if(rest->attribute() == nodeLabel.cstr())
                rest->append(id);
        };

        rest->forEachNodeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::NODEREST, RestrictionType::NODEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeEdgeLabelRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            ogdf::String edgeLabel = attr->attr()->labelEdge(edge);
            int id = attr->idByEdge(edge);
            if(rest->attribute() == edgeLabel.cstr())
                rest->append(id);
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNegatedNodeLabelRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        NodeFunction func = [&] (ogdf::node node)
        {
            ogdf::String nodeLabel = attr->attr()->labelNode(node);
            int id = attr->idByNode(node);
            if(rest->attribute() != nodeLabel.cstr())
                rest->append(id);
        };

        rest->forEachNodeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::NODEREST, RestrictionType::NODEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNegatedEdgeLabelRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            ogdf::String edgeLabel = attr->attr()->labelEdge(edge);
            int id = attr->idByEdge(edge);
            if(rest->attribute() != edgeLabel.cstr())
                rest->append(id);
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}

std::shared_ptr<GDSLAttributeRest> gdsl::makeWithoutNodesRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable& scope, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        std::shared_ptr<GDSLVariable> nodes = scope.at(rest->attribute());
        GDSLRestList restrictions = nodes->restrictions();

        if(restrictions.size() == 1)
        {
            std::shared_ptr<GDSLRestriction> excludeRest = restrictions[0];

            NodeFunction func = [&] (ogdf::node node)
            {
                int id = attr->idByNode(node);
                if(excludeRest->outputType() == RestrictionType::NODEREST)
                {
                    for(auto it = excludeRest->begin(); it < excludeRest->end(); ++it)
                        if(*it != id)
                            rest->append(id);
                }
                else if(excludeRest->outputType() == RestrictionType::EDGEREST)
                {
                    for(auto it = excludeRest->begin(); it < excludeRest->end(); ++it)
                    {
                        ogdf::edge edge = attr->edgeById(*it);
                        int target = attr->idByNode(edge->target());
                        int source = attr->idByNode(edge->source());

                        if(target != id && source != id)
                            rest->append(id);
                    }
                }
            };

            rest->forEachNodeDo(attr, func);
        }

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::OBJECTREST, RestrictionType::NODEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeWithoutEdgesRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable& scope, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        EdgeFunction func = [&] (ogdf::edge edge)
        {
            std::shared_ptr<GDSLVariable> edges = scope.at(rest->attribute());
            GDSLRestList restrictions = edges->restrictions();
            int id = attr->idByEdge(edge);

            if(restrictions.size() == 1)
            {
                std::shared_ptr<GDSLRestriction> excludeRest = restrictions[0];

                if(excludeRest->outputType() == RestrictionType::EDGEREST)
                {
                    for(auto it = excludeRest->begin(); it < excludeRest->end(); ++it)
                    {
                        if(*it != id)
                            rest->append(id);
                    }
                }
            }
        };

        rest->forEachEdgeDo(attr, func);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::EDGEREST, RestrictionType::EDGEREST, tree, scope);
}


std::shared_ptr<GDSLAttributeRest> gdsl::makeNameNullRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLAttributeRest> restwp, GDSLVarTable& scope, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid>&)
    {
        auto rest = restwp.lock();
        rest->clear();

        auto var = scope[rest->attribute()];
        if(var->restrictions().size() > 1)
            ; //Error

        auto varRest = var->restrictions()[0];
        for(auto it = varRest->begin(); it < varRest->end(); ++it)
            rest->append(*it);

        return true;
    };

    return makeRestriction<GDSLAttributeRest>(apply, RestrictionType::GENERAL, RestrictionType::GENERAL, tree, scope);
}
