#include "gdslintrest.hpp"

#include "../expressions/mathexpr.hpp"

using namespace gdsl;

bool GDSLIntRest::parseFromTree(pANTLR3_BASE_TREE tree, std::unordered_map<string, std::shared_ptr<GDSLVariable> > & scope)
{
    //TODO: Exception Handling
    //TODO: Parse more than closed intervals
    m_range = std::make_shared<GDSLClosedIntvl>();

    if(tree == nullptr)
    {
        m_range->append(0);
        m_range->append(0);

        return true;
    }

    // False means that the intvl is open;
    bool opening = false;
    bool closing = false;

    std::function<bool (pANTLR3_BASE_TREE)> parse = [&] (pANTLR3_BASE_TREE tree)
    {
        pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);

        switch(tok->type)
        {
            case DOTS:
            {
                m_contignous = true;
                return true;
            }
            case NUMOF_COLS:
            {
                m_maxColPos.push_back(m_range->length());
                m_range->append(-1);

                return true;
            }
            case NUMOF_ROWS:
            {
                m_maxRowPos.push_back(m_range->length());
                m_range->append(-1);

                return true;
            }
            case MATH:
            {
                m_mathPos.push_back(m_range->length());

                auto mathExpr = std::make_shared<MathExpr>(tree, scope);
                m_mathExpr.push_back(mathExpr);
                m_range->append(-1);

                return true;
            }
            case PARENS_OPEN:
            {
                if(tree->getChildCount(tree) > 0)
                {
                    if(antlrutils::foreachChildDo(tree, parse))
                        return true;
                    else
                        return false;
                }

                return true;
            }
            case PARENS_CLOSE:
            {
                return true;
            }
            case BRACK_OPEN:
            {
                if(tree->getChildCount(tree) > 0)
                {
                    if(antlrutils::foreachChildDo(tree, parse))
                        return true;
                    else
                        return false;
                }

                opening = true;
                return true;
            }
            case BRACK_CLOSE:
            {
                closing = true;
                return true;
            }
            default:
            {
               return true;
            }
        }

        return true;
    };

    pANTLR3_COMMON_TOKEN tok = tree->getToken(tree);
    if(tok && tok->type == MATH)
        parse(tree);
    else
        antlrutils::foreachChildDo(tree, parse);

    return true;
}

bool GDSLIntRest::isSubset(const std::shared_ptr<GDSLRestriction> &) const
{
    return true;
}

std::string GDSLIntRest::toString() const
{
    return m_range->toString();
}

int GDSLIntRest::intbegin() const
{
    return *m_range->begin();
}

int GDSLIntRest::intend() const
{
    return *m_range->end();
}



void GDSLIntRest::evalMathExpr(GDSLVarTable &scope, GDSLExpr::ConstantsMap& constants, std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
{
    for(int pos : m_mathPos)
    {
        auto exprs = m_mathExpr[pos]->eval(scope, constants, attr, grid);
        auto expr = exprs->at(0);

        if(exprs->isNumber())
            m_range->setAt(pos, (*exprs->begin())->getConstant());
        else
        {
            auto varCoeff = *expr->begin();
            auto var = varCoeff.second;
            auto rests = var->parent()->restrictions();

            double coeff = varCoeff.first;
            int varValue = *(rests[0]->begin() + var->id());
            int constant = (*exprs->begin())->getConstant();

            m_range->setAt(pos, coeff * varValue + constant);
        }
    }
}

std::shared_ptr<GDSLIntRest> gdsl::makeIntRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable& scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable& scope, ConstantsMap& constants,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        for(int pos : rest->maxColPos())
            rest->setAt(pos, grid->x() - 1);

        for(int pos : rest->maxRowPos())
            rest->setAt(pos, grid->y() - 1);

        for(auto expr : rest->mathExpr())
            expr->apply(attr, grid);

        rest->evalMathExpr(scope, constants, attr, grid);

        if(rest->contignous())
            rest->makeRangeContignous();

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::INTREST, RestrictionType::INTREST, tree, scope);
}

std::shared_ptr<GDSLIntRest> gdsl::makeMaxIntRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable& scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        if(grid->x() > grid->y())
            rest->setBeginEnd(0, grid->x() - 1);
        else
            rest->setBeginEnd(0, grid->y() - 1);

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::INTREST, RestrictionType::INTREST, tree, scope);
}

std::shared_ptr<GDSLIntRest> gdsl::makeRowRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable& scope, ConstantsMap& constants,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        for(int pos : rest->maxColPos())
            rest->setAt(pos, grid->x() - 1);

        for(int pos : rest->maxRowPos())
            rest->setAt(pos, grid->y() - 1);

        for(auto expr : rest->mathExpr())
            expr->apply(attr, grid);

        rest->evalMathExpr(scope, constants, attr, grid);

        if(rest->contignous())
            rest->makeRangeContignous();

        if(grid->y() < *rest->begin())
            throw OutOfBoundsException("Grid " + grid->toString() + " to small for RowRestriction " + rest->toString());

        if(grid->y() < *(rest->end() - 1))
            throw OutOfBoundsException("Grid " + grid->toString() + " to small for RowRestriction " + rest->toString());

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::ROWREST, RestrictionType::ROWREST, tree, scope);
}


std::shared_ptr<GDSLIntRest> gdsl::makeColumnRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable& scope, ConstantsMap& constants,
            std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        for(int pos : rest->maxColPos())
            rest->setAt(pos, grid->x() - 1);

        for(int pos : rest->maxRowPos())
            rest->setAt(pos, grid->y() - 1);

        for(auto expr : rest->mathExpr())
            expr->apply(attr, grid);

        rest->evalMathExpr(scope, constants, attr, grid);

        if(rest->contignous())
            rest->makeRangeContignous();

        if(grid->x() - 1 < *rest->begin())
            throw OutOfBoundsException("Grid " + grid->toString() + " to small for ColumnRestriction " + rest->toString());

        if(grid->x() - 1 < *(rest->end() - 1))
            throw OutOfBoundsException("Grid " + grid->toString() + " to small for ColumnRestriction " + rest->toString());

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::COLUMNREST, RestrictionType::COLUMNREST, tree, scope);
}


std::shared_ptr<GDSLIntRest> gdsl::makeMaxRowRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        rest->setBeginEnd(0, grid->y() - 1);

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::ROWREST, RestrictionType::ROWREST, tree, scope);
}


std::shared_ptr<GDSLIntRest> gdsl::makeMaxColumnRestriction(pANTLR3_BASE_TREE tree, GDSLVarTable &scope)
{
    auto apply = [] (std::weak_ptr<GDSLIntRest> restwp, GDSLVarTable&, ConstantsMap&,
            std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid> &grid)
    {
        auto rest = restwp.lock();

        rest->setBeginEnd(0, grid->x() - 1);

        return true;
    };

    return makeRestriction<GDSLIntRest>(apply, RestrictionType::COLUMNREST, RestrictionType::COLUMNREST, tree, scope);
}
