#ifndef gdslrestriction_hpp
#define gdslrestriction_hpp 1

#include <memory>
#include <vector>
#include <functional>
#include <unordered_map>
#include <string>

#include "../GDSLGrammarParser.h"
#include "../GDSLGrammarLexer.h"

#include "gdslrange.hpp"
#include "../grid.hpp"
#include "../randomaccessgraph.hpp"
#include "../antlrutilities.hpp"
#include "gdslclosedintvl.hpp"
#include "notanintegerexception.hpp"
#include "../outofboundsexception.hpp"

namespace gdsl
{
class GDSLVariable;
class GDSLRestriction;

using namespace std::placeholders;
enum class RestrictionType {GENERAL, ROWREST, COLUMNREST, OBJECTREST, NODEREST, EDGEREST, INTREST};

typedef std::vector< std::shared_ptr<GDSLRestriction> > GDSLRestList;
typedef std::unordered_map< std::string, std::shared_ptr<GDSLVariable> > GDSLVarTable;
typedef std::unordered_map<std::string, int> ConstantsMap;

template<class Restriction>
using ApplyFunctionUnbound = std::function<bool(std::weak_ptr<Restriction>& rest,
                                                GDSLVarTable& scope,
                                                ConstantsMap& constants,
                                                std::shared_ptr<RandomAccessGraph>& attr,
                                                std::shared_ptr<Grid>& grid)>;

typedef std::function<bool(GDSLVarTable& scope, ConstantsMap& constants, std::shared_ptr<RandomAccessGraph>& attr, std::shared_ptr<Grid>& grid)> ApplyFunction;

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType, pANTLR3_BASE_TREE tree, GDSLVarTable& scope);

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType, GDSLRangeIterator& it);

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType, std::string& varName);

class GDSLRestriction
{     
public:
    GDSLRestriction(RestrictionType inputType, RestrictionType outputType) {m_inputType = inputType; m_outputType = outputType; m_innerRest = nullptr;}
    GDSLRestriction(GDSLRangeIterator& it, RestrictionType inputType, RestrictionType outputType)
    {
        m_inputType = inputType;
        m_outputType = outputType;
        m_innerRest = nullptr;
        m_range = std::make_shared<GDSLClosedIntvl>(*it, *it);
    }

    virtual bool parseFromTree(pANTLR3_BASE_TREE, GDSLVarTable&) {return true;}
    virtual bool isSubset(const std::shared_ptr<GDSLRestriction>&) {return true;} //TODO: implement
    virtual std::shared_ptr<GDSLRestriction> singleRestAt(GDSLRangeIterator& it) const
    {
        auto apply = [] (std::weak_ptr<GDSLRestriction>&, GDSLVarTable&, ConstantsMap&,
                std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid>&)
        {
            return true;
        };

        return makeRestriction<GDSLRestriction>(apply, outputType(), outputType(), it);
    }

    virtual std::string toString() const {return m_range->toString();}

    virtual GDSLRangeIterator begin() {return m_range->begin();}
    virtual GDSLRangeIterator end() {return m_range->end();}

    virtual GDSLRangeConstIterator begin() const {return m_range->begin();}
    virtual GDSLRangeConstIterator end() const {return m_range->end();}

    virtual int length() const {return m_range->length();}
    virtual int pos(std::shared_ptr<GDSLRestriction> rest) const {return m_range->pos(*rest->begin(), m_range->begin(), m_range->end());}

    void append(int id) {m_range->append(id);}

    int asInteger() const
    {
        if(m_range->length() == 1)
            return *(m_range->begin());

        throw NotAnIntegerException("");
    }

    void setApply(std::shared_ptr<ApplyFunction> apply)
    {
        m_applyFunc = apply;
    }

    std::shared_ptr<ApplyFunction> getApply()
    {
        return m_applyFunc;
    }

    virtual bool apply(GDSLVarTable& scope, ConstantsMap& constants,std::shared_ptr<RandomAccessGraph>& attr, std::shared_ptr<Grid>& grid)
    {
        if(m_innerRest != nullptr)
            m_innerRest->apply(scope, constants, attr, grid);

        return (*m_applyFunc)(scope, constants, attr, grid);
    }

    void clear() {m_range->clear();}

    RestrictionType inputType() const {return m_inputType;}
    RestrictionType outputType() const {return m_outputType;}

    std::shared_ptr<GDSLRestriction> innerRestriction() const {return m_innerRest;}
    void setInnerRestriction(std::shared_ptr<GDSLRestriction> innerRest) {m_innerRest = innerRest;}

    virtual void evalMathExpr(GDSLVarTable&, ConstantsMap&, std::shared_ptr<RandomAccessGraph>&, std::shared_ptr<Grid>&) {}

    void forEachEdgeDo(std::shared_ptr<RandomAccessGraph> &attr, std::function<void (ogdf::edge edge)>& func)
    {
        if(innerRestriction() == nullptr)
        {
            ogdf::edge edge = attr->attr()->constGraph().firstEdge();
            ogdf::edge lastEdge = attr->attr()->constGraph().lastEdge();

            if(edge)
            {
                func(edge);

                while(edge != lastEdge)
                {
                    edge = edge->succ();
                    func(edge);
                }
            }
        }
        else
            for(GDSLRangeIterator it = innerRestriction()->begin(); it < innerRestriction()->end(); ++it)
            {
                auto edge = attr->edgeById(*it);
                func(edge);
            }
    }

    void forEachNodeDo(std::shared_ptr<RandomAccessGraph> &attr, std::function<void (ogdf::node node)>& func)
    {
        if(innerRestriction() == nullptr)
        {
            ogdf::node node = attr->attr()->constGraph().firstNode();
            ogdf::node lastNode = attr->attr()->constGraph().lastNode();

            if(node)
            {
                func(node);

                while(node != lastNode)
                {
                    node = node->succ();
                    func(node);

                }
            }
        }
        else
            for(GDSLRangeIterator it = innerRestriction()->begin(); it < innerRestriction()->end(); ++it)
            {
                auto node = attr->nodeById(*it);
                func(node);
            }
    }

    void applyInnerRest(GDSLVarTable &scope, ConstantsMap& constants, std::shared_ptr<RandomAccessGraph> &attr, std::shared_ptr<Grid> &grid)
    {
        if(m_innerRest != nullptr)
            m_innerRest->apply(scope, constants, attr, grid);
    }

    bool isInInnerRest(int id)
    {
        return (m_innerRest == nullptr || std::find(m_innerRest->begin(), m_innerRest->end(), id) != m_innerRest->end());
    }

protected:
    std::shared_ptr<GDSLRestriction> m_innerRest;
    std::shared_ptr<GDSLRange> m_range;
    std::shared_ptr<ApplyFunction> m_applyFunc;
    RestrictionType m_inputType = RestrictionType::GENERAL;
    RestrictionType m_outputType = RestrictionType::GENERAL;
};

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType, pANTLR3_BASE_TREE tree, GDSLVarTable& scope)
{
    std::shared_ptr<Restriction> rest = std::make_shared<Restriction>(tree, scope, inputType, outputType);
    std::weak_ptr<Restriction> restWp(rest);

    auto applyBounded = std::bind(apply, restWp, _1, _2, _3, _4);
    auto applyBoundedFunction = std::make_shared<ApplyFunction>(applyBounded);
    rest->setApply(applyBoundedFunction);

    return rest;
}

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType, GDSLRangeIterator& it)
{
    std::shared_ptr<Restriction> rest = std::make_shared<Restriction>(it, inputType, outputType);
    std::weak_ptr<Restriction> restWp(rest);

    auto applyBounded = std::bind(apply, restWp, _1, _2, _3, _4);
    auto applyBoundedFunction = std::make_shared<ApplyFunction>(applyBounded);
    rest->setApply(applyBoundedFunction);

    return rest;
}

template<class Restriction>
std::shared_ptr<Restriction> makeRestriction(ApplyFunctionUnbound<Restriction> apply, RestrictionType inputType, RestrictionType outputType)
{
    std::shared_ptr<Restriction> rest = std::make_shared<Restriction>(inputType, outputType);
    std::weak_ptr<Restriction> restWp(rest);

    auto applyBounded = std::bind(apply, restWp, _1, _2, _3, _4);
    auto applyBoundedFunction = std::make_shared<ApplyFunction>(applyBounded);
    rest->setApply(applyBoundedFunction);

    return rest;
}

}
#endif // gdslrestriction_hpp
