#include "gdslpreprocessor.hpp"

#include "preprocessorutilities.hpp"

using namespace gdsl;

const std::string Preprocessor::STARTSYM = "%";

const std::string Preprocessor::SETNAME = "name";
const std::string Preprocessor::SET = "set";
const std::string Preprocessor::EXEC = "execute";
const std::string Preprocessor::EXEC_FOR = "for";
const std::string Preprocessor::USE = "use";

Preprocessor::Preprocessor(const std::string& fileName, const std::string &program, const std::string &makroDir, bool followDependencies) :
    m_origProgram(program),
    m_fileName(fileName),
    m_makroDir(makroDir),
    m_followDependencies(followDependencies)
{
    if(!m_fileName.empty())
        m_dependencies.addNode(m_fileName);
}

void Preprocessor::parsePpStatements()
{
    std::stringstream programStream(m_origProgram);
    std::string line;
    std::vector<std::string> ppStmts;

    int lineNumber = 0;
    while(std::getline(programStream, line))
    {
        if(std::equal(STARTSYM.begin(), STARTSYM.end(), line.begin()))
            ppStmts.push_back(line);
        else
            m_programWithoutPp += line + "\n";

        ++lineNumber;
    }

    std::sort(ppStmts.begin(), ppStmts.end(), [] (const std::string& first, const std::string&)
                                                    {
                                                        auto searchString = STARTSYM + USE;

                                                        if(first.substr(0, searchString.length()) == searchString)
                                                            return false;

                                                        return true;
                                                    });

    for(auto stmt : ppStmts)
    {
        stmt.erase(stmt.begin(), stmt.begin() + STARTSYM.length());
        parseLine(stmt);
    }

    findHelpFile();
}

bool Preprocessor::applyPpStatements()
{
    m_processedProgram = m_programWithoutPp;
    m_dependencies.setProgram(fileName(), m_processedProgram);
    return true;
}

void Preprocessor::setIntVar(std::string &name, int value)
{
    std::string valueStr(std::to_string(value));
    setVar(name, valueStr);
}

void Preprocessor::setBoolVar(std::string &name, bool value)
{
    std::string valueStr(std::to_string(value));
    setVar(name, valueStr);
}

std::unordered_map<std::string, int> Preprocessor::constants() const
{
    std::unordered_map<std::string, int> constants;

    for(auto stmt : m_valueStmts)
        constants[stmt->name()] = stmt->value();

    return constants;
}

bool Preprocessor::next()
{
    bool changed = false;

    for(auto stmt : m_valueStmts)
        changed |= stmt->next();

    return changed;
}

void Preprocessor::parseLine(std::string &ppLine)
{
    PPUtils::collapseWhitespaces(ppLine);
    std::string identifier = PPUtils::pop(ppLine);

    if(identifier == SETNAME)
        m_name = PPUtils::getName(ppLine);
    else if(identifier == SET || (identifier == EXEC && PPUtils::pop(ppLine) == EXEC_FOR))
    {
        auto stmt = ValueStmt::parseLine(ppLine);
        m_valueStmts.push_back(stmt);
    }
    else if(identifier == USE)
    {
        if(m_followDependencies)
        {
            UseStmt stmt(ppLine, m_makroDir);
            mergeDependencies(stmt);

            if(!fileName().empty())
                m_dependencies.addDependency(fileName(), stmt.fileName());
        }
        else
            ppLine.clear();
    }
    else
        throw PreprocessorException("Unknown Preprocessor Statement '" + ppLine + "'");
}

void Preprocessor::setVar(std::string &name, std::string &valueStr)
{
    for(auto stmt : m_valueStmts)
        if(stmt->name() == name)
            stmt->setValue(valueStr);
}

std::string Preprocessor::startForVar(std::string name) const
{
    for(auto stmt : m_valueStmts)
        if(stmt->name() == name)
            return stmt->valueStartStr();
}

std::string Preprocessor::endForVar(std::string name) const
{
    for(auto stmt : m_valueStmts)
        if(stmt->name() == name)
            return stmt->valueEndStr();
}

std::vector<std::string> Preprocessor::getNames(ValueStmt::RET_TYPE type) const
{
    std::vector<std::string> names;

    for(auto stmt : m_valueStmts)
        if(stmt->returnType() == type)
            names.push_back(stmt->name());

    return names;
}

void Preprocessor::mergeDependencies(const UseStmt &stmt)
{
    auto dependencyGraph = stmt.dependencies();

    for(auto node : dependencyGraph.nodes())
    {
        if(!m_dependencies.containsNode(node))
        {
            m_dependencies.addNode(node);
            m_dependencies.setProgram(node, dependencyGraph.programForNode(node));
        }

        for(auto dep : dependencyGraph.dependenciesForNode(node))
            m_dependencies.addDependency(node, dep);
    }
}

void Preprocessor::findHelpFile()
{   
    boost::filesystem::path makroFile(m_fileName);
    boost::filesystem::path makroHelpFile(makroFile.parent_path() / makroFile.stem() += ".html");
    boost::filesystem::path makroDirPath(m_makroDir);

    boost::filesystem::recursive_directory_iterator endIt;
    for(boost::filesystem::recursive_directory_iterator dirIt(makroDirPath); dirIt != endIt; ++dirIt)
    {
        auto res = dirIt->path().string().find(makroHelpFile.string());
        if(res != std::string::npos)
        {
            std::ifstream makroFileFS(dirIt->path().string());
            std::stringstream ss;

            ss << makroFileFS.rdbuf();
            m_helpText = ss.str();
        }
    }
}


bool MakroPreprocessor::applyPpStatements()
{
    m_processedProgram = m_programWithoutPp;

    for(auto constant : constants())
    {
        std::string value = std::to_string(constant.second);
        std::string constantStr = constant.first;
        PPUtils::replaceAll(constantStr, value, m_processedProgram);
    }

    m_dependencies.setProgram(fileName(), m_processedProgram);

    return true;
}


bool GDSLPreprocessor::applyPpStatements()
{
    auto sorted = m_dependencies.topoSort();
    std::string wholeProgram;

    for(auto node : sorted)
        wholeProgram.append(m_dependencies.programForNode(node));

    wholeProgram.append(m_programWithoutPp);

    m_processedProgram = wholeProgram;
    m_dependencies.setProgram(fileName(), m_processedProgram);

    return true;
}
