#ifndef GDSLPREPROCESSOR_HPP
#define GDSLPREPROCESSOR_HPP

#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <vector>
#include <algorithm>
#include <functional>
#include <memory>
#include <unordered_map>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/progress.hpp>

#include "usestmt.hpp"
#include "valuestmt.hpp"
#include "preprocessorutilities.hpp"
#include "../parser/restrictions/gdslclosedintvl.hpp"

namespace gdsl
{
class Preprocessor
{
    static const std::string STARTSYM;

    static const std::string SETNAME;
    static const std::string SET;
    static const std::string USE;
    static const std::string EXEC;
    static const std::string EXEC_FOR;

    friend class MakroPreprocessor;
    friend class GDSLPreprocessor;

public:
    Preprocessor(bool followDependencies = true) : m_followDependencies(followDependencies) {}
    Preprocessor(const std::string &fileName, const std::string &program, const std::string &makroDir, bool followDependencies = true);

    virtual void parsePpStatements();
    virtual bool applyPpStatements();

    void setIntVar(std::string& name, int value);
    void setBoolVar(std::string& name, bool value);
    void setVar(std::string& name, std::string& valueStr);

    std::vector<std::string> intVars() const {return getNames(ValueStmt::RET_TYPE::PPINT);}
    std::vector<std::string> boolVars() const {return getNames(ValueStmt::RET_TYPE::PPBOOL);}
    std::vector<std::string> constVars() const {return getNames(ValueStmt::RET_TYPE::PPCONST);}

    std::string startForVar(std::string name) const;
    std::string endForVar(std::string name) const;

    std::string programWithoutPpStmts() const {return m_programWithoutPp;}
    std::string preprocessedProgram() const {return m_processedProgram;}
    std::string help() const {return m_helpText;}
    std::string name() const {return (m_name.empty() ? m_fileName : m_name);}
    std::string fileName() const {return m_fileName;}
    const DependencyGraph& dependencies() const {return m_dependencies;}

    void setFilename(const std::string& fileName) {m_fileName = fileName;}
    void setProgram(const std::string& program) {m_origProgram = program;}
    void setMakroDir(const std::string& makroDir) {m_makroDir = makroDir;}

    std::unordered_map<std::string, int> constants() const;
    bool next();

private:
    void parseLine(std::string& ppLine);
    void findHelpFile();
    std::vector<std::string> getNames(ValueStmt::RET_TYPE type) const;
    void mergeDependencies(const UseStmt& stmt);

    std::string m_origProgram;
    std::string m_programWithoutPp;
    std::string m_processedProgram;

    std::string m_fileName;
    std::string m_name;
    std::string m_makroDir;
    bool m_followDependencies;

    std::vector<std::shared_ptr<ValueStmt>> m_valueStmts;
    DependencyGraph m_dependencies;

    std::string m_helpText;
};

class GDSLPreprocessor : public Preprocessor
{
public:
    GDSLPreprocessor(bool followDependencies = true) : Preprocessor(followDependencies) {}
    GDSLPreprocessor(std::string &fileName, std::string &program, std::string &makroDir, bool followDependencies = true) :
        Preprocessor(fileName, program, makroDir, followDependencies)
    {}

    virtual bool applyPpStatements();
};

class MakroPreprocessor : public Preprocessor
{
public:
    MakroPreprocessor(bool followDependencies = true) : Preprocessor(followDependencies) {}
    MakroPreprocessor(std::string &fileName, std::string &program, std::string &makroDir, bool followDependencies = true) :
        Preprocessor(fileName, program, makroDir, followDependencies)
    {}

    virtual bool applyPpStatements();
};
}

#endif // GDSLPREPROCESSOR_HPP
