#ifndef PREPROCESSORSUTILITIES_HPP
#define PREPROCESSORSUTILITIES_HPP

#include <string>
#include <memory>
#include <iostream>
#include <algorithm>
#include <sstream>
#include <unordered_map>

#include "../parser/gdslexception.hpp"

namespace gdsl
{

class UseStmt;

class PreprocessorException : public GDSLException
{
public:
    PreprocessorException(std::string error) noexcept : GDSLException(error)  {}
    ~PreprocessorException() {}

    const char* what() const noexcept;
};

struct PPUtils
{
    static void collapseWhitespaces(std::string& ppLine);
    static void stripWhitespaces(std::string& str);
    static std::string pop(std::string& ppLine, char delimiter = ' ', bool deleteDelimiter = true);
    static char popFirst(std::string& ppLine);
    static bool isName(const std::string& name);
    static bool isNumber(const std::string& number);
    static int toInt(const std::string& number);
    static bool isBool(const std::string& boolean);
    static bool toBool(const std::string& boolean);
    static void replaceAll(std::string& oldStr, std::string& newStr, std::string& program);
    static std::string getName(std::string& ppLine);
};

class DependencyGraph
{
public:
    DependencyGraph() {}

    void addNode(const std::string id)
    {
        m_dependencies[id] = std::vector<std::string>();

        if(m_parents.find(id) == m_parents.end())
            m_parents[id] = std::vector<std::string>();

        m_programs[id] = "";
    }

    void setProgram(const std::string id, const std::string program) {m_programs[id] = program;}

    void addDependency(const std::string id, const std::string dependency)
    {
        auto currentDependencies = m_dependencies.at(id);

        if(std::find(currentDependencies.begin(), currentDependencies.end(), dependency) == currentDependencies.end())
        {
            m_dependencies[id].push_back(dependency);
            m_parents[dependency].push_back(id);
        }
    }

    std::vector<std::string> nodes() const
    {
        std::vector<std::string> keys;

        for(auto dep : m_dependencies)
            keys.push_back(dep.first);

        return keys;
    }

    std::vector<std::string> topoSort() const
    {
        std::vector<std::string> sorted;
        std::vector<std::string> next;
        std::unordered_map<std::string, bool> inserted;
        auto nodeKeys = nodes();

        for(auto node : nodeKeys)
            inserted[node] = false;

        for(auto node : nodeKeys)
        {
            if(m_parents.at(node).empty())
            {
                if(!inserted[node])
                {
                    for(auto dep : m_dependencies.at(node))
                        next.push_back(dep);

                    inserted[node] = true;
                    sorted.push_back(node);
                }
            }
        }

        while(!next.empty())
        {
            std::vector<std::string> newNext;

            for(auto node : next)
            {
                if(!inserted[node])
                {                    for(auto dep : m_dependencies.at(node))
                        newNext.push_back(dep);

                    inserted[node] = true;
                    sorted.push_back(node);
                }
            }

            next.swap(newNext);
        }

        std::reverse(sorted.begin(), sorted.end());

        return sorted;
    }

    bool containsNode(const std::string& node) const {return m_dependencies.find(node) == m_dependencies.end() ? false : true;}
    bool empty() const {return m_dependencies.empty() && m_programs.empty();}

    std::vector<std::string> dependenciesForNode(const std::string& node) const {return m_dependencies.at(node);}
    std::vector<std::string> parentsForNode(const std::string& node) const {return m_parents.at(node);}
    std::string programForNode(const std::string& node) const {return m_programs.at(node);}

private:
    std::unordered_map<std::string, std::vector<std::string>> m_dependencies;
    std::unordered_map<std::string, std::string> m_programs;

    std::unordered_map<std::string, std::vector<std::string>> m_parents;
};
}

#endif // PREPROCESSORSTMT_HPP
