#include "usestmt.hpp"

#include "gdslpreprocessor.hpp"

using namespace gdsl;

const std::string UseStmt::MAKRO = "makro";
const std::string UseStmt::PRE_WITH = "with";
const char UseStmt::PRE_COMMA = ',';
const char UseStmt::PRE_ASSIGN = '=';

UseStmt::UseStmt(std::string &ppLine, const std::string &makroDir) :
    m_makroDir(makroDir)
{
    auto token = PPUtils::pop(ppLine);

    if(token != MAKRO)
        throw PreprocessorException("Expected makro, got " + token);

    auto file = PPUtils::pop(ppLine);

    auto depPath = findGDSLFile(file);
    auto program = readGDSLFile(depPath);

    m_preprocessor = std::make_shared<MakroPreprocessor>(file, program, m_makroDir);
    m_preprocessor->parsePpStatements();

    if(!ppLine.empty())
        parseWithStmts(ppLine);

    m_preprocessor->applyPpStatements();
}

std::string UseStmt::program() const
{
    return m_preprocessor->preprocessedProgram();
}

std::string UseStmt::name() const
{
    return m_preprocessor->name();
}

std::string UseStmt::fileName() const
{
    return m_preprocessor->fileName();
}

const DependencyGraph &UseStmt::dependencies() const
{
    return m_preprocessor->dependencies();
}

boost::filesystem::path UseStmt::findGDSLFile(const std::string &path)
{
    fs::path makroDirPath(m_makroDir);
    fs::recursive_directory_iterator dirIt(makroDirPath);
    fs::recursive_directory_iterator endIt;

    for(; dirIt != endIt; ++dirIt)
    {
        std::string makroPath = dirIt->path().string();
        auto res = makroPath.find(path);

        if(res != std::string::npos)
            return dirIt->path();
    }

    throw PreprocessorException("Could not find macro " + path);
}

std::string UseStmt::readGDSLFile(const fs::path &path)
{
    try
    {
        std::ifstream makroFileStream(path.string());
        std::stringstream ss;
        ss << makroFileStream.rdbuf();
        return ss.str();
    }
    catch(std::ios_base::failure e)
    {
        throw PreprocessorException("Could not read macro file " + path.string());
    }
}

void UseStmt::parseWithStmts(std::string &ppLine)
{
    auto token = PPUtils::pop(ppLine);
    if(token != PRE_WITH)
        throw PreprocessorException("Expected with, got " + token);

    while(!ppLine.empty())
    {
        auto name = PPUtils::pop(ppLine, PRE_ASSIGN, false);
        PPUtils::stripWhitespaces(name);
        if(!PPUtils::isName(name))
            throw PreprocessorException(name + " is not a valid name.");

        auto assign = PPUtils::popFirst(ppLine);
        if(assign != PRE_ASSIGN)
            throw PreprocessorException("Expected =, got " + assign);

        auto value = PPUtils::pop(ppLine, PRE_COMMA);
        PPUtils::stripWhitespaces(value);
        if(!PPUtils::isNumber(value) && !PPUtils::isBool(value))
            throw PreprocessorException(value + " is neither a number nor a boolean.");

        m_preprocessor->setVar(name, value);
    }
}
