#include "gdsleditor.h"

GDSLEditor::GDSLEditor(QWidget *parent) :
    QPlainTextEdit(parent)
{
    m_lineNumberArea = new LineNumberArea(this);

    connect(this, SIGNAL(blockCountChanged(int)), this, SLOT(updateLineNumberAreaWidth(int)));
    connect(this, SIGNAL(updateRequest(QRect,int)), this, SLOT(updateLineNumberArea(QRect,int)));
    connect(this, SIGNAL(cursorPositionChanged()), this, SLOT(highlightCurrentLine()));
    connect(this, SIGNAL(textChanged()), this, SLOT(complete()));
    connect(this, SIGNAL(textChanged()), this, SLOT(updateVariableCompletion()));

    updateLineNumberAreaWidth(0);
    highlightCurrentLine();

    m_standardCompletions   << "nodes"
                            << "edges"
                            << "rows"
                            << "cols"
                            << "objects";

    m_model = new QStringListModel(this);
    m_model->setStringList(m_standardCompletions);

    m_completer = new QCompleter;
    m_completer->setModel(m_model);
    m_completer->setWidget(this);

    connect(m_completer, SIGNAL(activated(QString)), this, SLOT(complete(QString)));
}

GDSLEditor::~GDSLEditor()
{
    delete m_completer;
    delete m_model;
}

void GDSLEditor::lineNumberAreaPaintEvent(QPaintEvent *event)
{
    QPainter painter(m_lineNumberArea);
    painter.fillRect(event->rect(), Qt::lightGray);

    QTextBlock block = firstVisibleBlock();
    int blockNumber = block.blockNumber();
    int top = (int) blockBoundingGeometry(block).translated(contentOffset()).top();
    int bottom = top + (int) blockBoundingRect(block).height();

    while (block.isValid() && top <= event->rect().bottom()) {
        if (block.isVisible() && bottom >= event->rect().top()) {
            QString number = QString::number(blockNumber + 1);
            painter.setPen(Qt::black);
            painter.drawText(0, top, m_lineNumberArea->width(), fontMetrics().height(),
                             Qt::AlignRight, number);
        }

        block = block.next();
        top = bottom;
        bottom = top + (int) blockBoundingRect(block).height();
        ++blockNumber;
    }
}

int GDSLEditor::lineNumberAreaWidth()
{
    int digits = 1;
    int max = qMax(1, blockCount());
    while (max >= 10) {
        max /= 10;
        ++digits;
    }

    int space = 3 + fontMetrics().width(QLatin1Char('9')) * digits;

    return space;
}

void GDSLEditor::resizeEvent(QResizeEvent *event)
{
    QPlainTextEdit::resizeEvent(event);

    QRect cr = contentsRect();
    m_lineNumberArea->setGeometry(QRect(cr.left(), cr.top(), lineNumberAreaWidth(), cr.height()));
}

void GDSLEditor::keyPressEvent(QKeyEvent *e)
{
    if(m_completer->popup()->isVisible())
    {
        if(e->key() == Qt::Key_Return && !e->modifiers())
        {
            m_completer->setCurrentRow(m_completer->popup()->currentIndex().row());
            complete(m_completer->currentCompletion());
            e->accept();
            return;
        }
    }

    if(e->key() == Qt::Key_Return && currentLineIsFor() && textCursor().atBlockEnd())
    {
        textCursor().insertText("\n    \nend");
        moveCursor(QTextCursor::PreviousBlock);
        moveCursor(QTextCursor::EndOfLine);
        return;
    }

    QPlainTextEdit::keyPressEvent(e);
}

void GDSLEditor::highlightCurrentLine()
{
    QList<QTextEdit::ExtraSelection> extraSelections;

    if (!isReadOnly())
    {
        QTextEdit::ExtraSelection selection;

        QColor lineColor = QColor(Qt::blue).lighter(190);

        selection.format.setBackground(lineColor);
        selection.format.setProperty(QTextFormat::FullWidthSelection, true);
        selection.cursor = textCursor();
        selection.cursor.clearSelection();
        extraSelections.append(selection);
    }

    setExtraSelections(extraSelections);
}

void GDSLEditor::updateLineNumberArea(const QRect& rect, int dy)
{
    if (dy)
        m_lineNumberArea->scroll(0, dy);
    else
        m_lineNumberArea->update(0, rect.y(), m_lineNumberArea->width(), rect.height());

    if (rect.contains(viewport()->rect()))
        updateLineNumberAreaWidth(0);
}

void GDSLEditor::complete()
{
    QTextCursor cursor = textCursor();
    cursor.select(QTextCursor::WordUnderCursor);

    QString currentWord = cursor.selectedText();
    m_completer->setCompletionPrefix(currentWord);

    if(!(currentWord.isEmpty() || m_completer->completionCount() == 0))
    {
        if(m_completer->completionCount() > 0)
        {
            QRect cursorRectangle = cursorRect();
            QRect popupRect(cursorRectangle.x(), cursorRectangle.y(), cursorRectangle.width() + 100, cursorRectangle.height() + 4);
            m_completer->complete(popupRect);
        }
    }
    else if(m_completer->popup()->isVisible())
        m_completer->popup()->close();

}

void GDSLEditor::complete(QString str)
{
    QTextCursor cursor = textCursor();
    cursor.select(QTextCursor::WordUnderCursor);
    cursor.insertText(str);
    m_completer->popup()->close();
}

void GDSLEditor::updateVariableCompletion()
{
    QRegExp exp("def\\s+([a-zA-Z]\\w*)\\([\\(\\)\\w\\s,]*\\)");
    QTextBlock block = document()->begin();
    QString line;
    QString name;
    QStringList completions = m_standardCompletions;

    while(block.isValid())
    {
        line = block.text();
        if(line.startsWith("def"))
        {
            int index = line.indexOf(exp);

            if(index != -1)
            {
                name = exp.cap(1);
                completions << name;
            }
        }

        block = block.next();
    }

    m_model->setStringList(completions);
}

bool GDSLEditor::currentLineIsFor() const
{
    QString line = textCursor().block().text();
    QRegExp forexp("for .+ in .+ do");

    if(forexp.indexIn(line) != -1)
        return true;

    return false;
}
