#include "gdslmakrochooser.hpp"
#include "ui_gdslmakrochooser.h"

#include "settingsdialog.hpp"

GDSLMakroChooser::GDSLMakroChooser(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::GDSLMakroChooser)
{
    ui->setupUi(this);
    m_currentMakro = nullptr;
    m_emptyLabel = new QLabel(this);

    QSettings settings;
    m_makroPath = settings.value(SettingsKeys::MAKRO_PATH).toString();

    ui->availableMakrosLayout->setStretchFactor(ui->availableMakros, 7);
    ui->availableMakrosLayout->addWidget(m_emptyLabel, 9);
}

GDSLMakroChooser::~GDSLMakroChooser()
{
    delete ui;
}

void GDSLMakroChooser::loadMakroList(const QString path, QTreeWidgetItem* parent)
{
    QDir makroDir(m_makroPath);
    QDir currentDir;
    if(path != "")
        currentDir = QDir(path);
    else
        currentDir = QDir(m_makroPath);

    QFileInfoList fileInfos = currentDir.entryInfoList(currentDir.filter() | QDir::NoDotAndDotDot);

    for(QFileInfo info : fileInfos)
    {
        if(info.isDir())
        {
            QTreeWidgetItem* item = new QTreeWidgetItem();
            item->setText(0, info.fileName().replace("_", " "));

            loadMakroList(info.absoluteFilePath(), item);

            if(parent != nullptr)
                parent->addChild(item);
            else
                ui->availableMakros->addTopLevelItem(item);
        }
        else if(info.isFile() && info.suffix() == "gdsl")
        {
            QFile makroFile(info.filePath());
            if(!makroFile.open(QIODevice::ReadOnly))
                continue;

            QString makroString(makroFile.readAll());
            QTreeWidgetItem* item = new QTreeWidgetItem();
            Makro* makro = new Makro(makroString, m_makroPath, makroDir.relativeFilePath(info.filePath()), item, this);
            makro->hide();
            ui->availableMakrosLayout->addWidget(makro, 9);
            m_makros.append(makro);
            connect(ui->availableMakros, SIGNAL(itemClicked(QTreeWidgetItem*, int)), this, SLOT(makroClicked(QTreeWidgetItem*, int)));
            connect(ui->availableMakros, SIGNAL(itemChanged(QTreeWidgetItem*,int)), this, SLOT(makroChanged(QTreeWidgetItem*,int)));

            item->setText(0, makro->displayName());
            item->setFlags(item->flags() | Qt::ItemIsUserCheckable);
            item->setCheckState(0, Qt::Unchecked);
            item->setData(0, Qt::UserRole, m_makros.size() - 1);

            if(parent != nullptr)
                parent->addChild(item);
            else
                ui->availableMakros->addTopLevelItem(item);
        }
    }
}

void GDSLMakroChooser::loadFromDependencyGraph(const gdsl::DependencyGraph &depGraph)
{
    for(auto dep : depGraph.nodes())
        for(auto makro : m_makros)
            if(makro->fileName() == QString::fromStdString(dep))
            {
                makro->treeItem()->setCheckState(0, Qt::Checked);
                makroChanged(makro->treeItem(), 0);
            }
}

std::vector<std::string> GDSLMakroChooser::processedMakros() const
{
    std::vector<std::string> processedMakros;

    for(int i = 0; i < m_checkedItems.count(); ++i)
    {
        QTreeWidgetItem* item = m_checkedItems.at(i);
        QString program = m_makros.at(item->data(0, Qt::UserRole).toInt())->useString();
        processedMakros.push_back(program.toStdString());
    }

    return processedMakros;
}

void GDSLMakroChooser::makroClicked(QTreeWidgetItem *item, int)
{
    if(item->childCount() == 0)
    {
        if(m_currentMakro)
            m_currentMakro->hide();

        m_emptyLabel->hide();

        m_currentMakro = m_makros.at(item->data(0, Qt::UserRole).toInt());
        m_currentMakro->show();
    }
}

void GDSLMakroChooser::makroChanged(QTreeWidgetItem *item, int column)
{
    if(item->checkState(column) == Qt::Checked && m_checkedItems.indexOf(item) == -1)
    {
        m_checkedItems.append(item);

        Makro* makro = m_makros.at(item->data(0, Qt::UserRole).toInt());
        makro->setUserChecked(true);
        for(auto dep : makro->dependencies())
            for(auto makro : m_makros)
                if(makro->fileName() == dep)
                {
                    if(m_checkedItems.indexOf(makro->treeItem()) == -1)
                        m_checkedItems.append(makro->treeItem());

                    makro->treeItem()->setCheckState(0, Qt::Checked);
                    makro->treeItem()->setDisabled(true);
                    makro->increaseUseCounter();
                }
    }
    else if(item->checkState(column) == Qt::Unchecked && m_checkedItems.indexOf(item) != -1)
    {
        m_checkedItems.remove(m_checkedItems.indexOf(item));

        Makro* makro = m_makros.at(item->data(0, Qt::UserRole).toInt());
        makro->setUserChecked(false);
        for(auto dep : makro->dependencies())
            for(auto makro : m_makros)
                if(makro->fileName() == dep)
                {
                    makro->decreaseUseCounter();

                    if(makro->useCounter() == 0)
                    {
                        if(!makro->checkedByUser())
                        {
                            m_checkedItems.remove(m_checkedItems.indexOf(makro->treeItem()));
                            makro->treeItem()->setCheckState(0, Qt::Unchecked);
                        }

                        makro->treeItem()->setDisabled(false);
                    }
                }
    }
}


Makro::Makro(QString makro, QString makroDir, QString makroFile, QTreeWidgetItem* correspondingItem, QWidget *parent) :
    QWidget(parent),
    m_makro(makro),
    m_makroDir(makroDir),
    m_fileName(makroFile),
    m_item(correspondingItem)
{
    QVBoxLayout* layout = new QVBoxLayout(this);
    layout->setMargin(0);

    auto stdMakro = m_makro.toStdString();
    auto makroStdDir = m_makroDir.toStdString();
    auto stdFileName = m_fileName.toStdString();

    gdsl::GDSLPreprocessor preprocessor(stdFileName, stdMakro, makroStdDir);
    preprocessor.parsePpStatements();

    m_name = QString::fromStdString(preprocessor.name());

    QLayout* variableLayout = new QVBoxLayout;
    variableLayout->setContentsMargins(-1, 0, -1, 0);
    variableLayout->setSpacing(0);

    for(auto key : preprocessor.intVars())
    {
        QString label(QString::fromStdString(key));
        IntEdit* intEdit = new IntEdit(label, this);
        variableLayout->addWidget(intEdit);
        m_intEdits.append(intEdit);
    }

    for(auto key : preprocessor.boolVars())
    {
        QString label(QString::fromStdString(key));
        BoolEdit* boolEdit = new BoolEdit(label, this);
        variableLayout->addWidget(boolEdit);
        m_boolEdits.append(boolEdit);
    }

    if(m_intEdits.size() > 0 || m_boolEdits.size() > 0)
    {
        QGroupBox* variableBox = new QGroupBox("Variables", this);
        variableBox->setLayout(variableLayout);
        layout->addWidget(variableBox);
    }

    if(preprocessor.help() != "")
    {
        QSettings settings;
        QFile cssFile(settings.value("paths/helpstyle").toString());

        QTextEdit* help = new QTextEdit(this);
        help->setReadOnly(true);

        if(cssFile.open(QIODevice::ReadOnly))
            help->document()->setDefaultStyleSheet(cssFile.readAll());

        help->setText(QString::fromStdString(preprocessor.help()));
        help->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Expanding);
        help->setFixedHeight(450);
        layout->addWidget(help);
    }

    layout->addStretch();

    gdsl::DependencyGraph depGraph = preprocessor.dependencies();
    for(auto dep : depGraph.nodes())
        if(QString::fromStdString(dep) != m_fileName)
            m_dependencies.append(QString::fromStdString(dep));
}

QString Makro::useString()
{
    QString useString;
    useString.append(tr("%use makro %1 ").arg(m_fileName));

    if(m_intEdits.size() > 0 || m_boolEdits.size() > 0)
    {
        useString.append("with ");

        for(IntEdit* intEdit : m_intEdits)
        {
            int value = intEdit->value();
            QString name = intEdit->label();

            useString.append(tr("%1 = %2").arg(name).arg(value));

            if(m_intEdits.last() != intEdit || m_boolEdits.size() > 0)
                useString.append(",");
        }

        for(BoolEdit* boolEdit : m_boolEdits)
        {
            int value = boolEdit->value();
            QString name = boolEdit->label();

            useString.append(tr("%1 = %2").arg(name).arg(value));

            if(m_boolEdits.last() != boolEdit)
                useString.append(",");
        }
    }

    return useString;
}
