#include "gdsluiprogram.hpp"

#include "../GDSL/parser/parseexception.hpp"
#include "../GDSL/parser/outofboundsexception.hpp"
#include "../GDSL/parser/expressions/gdslprogram.hpp"
#include "../GDSL/preprocessor/gdslpreprocessor.hpp"
#include "../GDSL/preprocessor/preprocessorutilities.hpp"

void GDSLuiProgram::preprocess(string &name, string &program, std::string &makroDir)
{
    try
    {
        m_preprocessor = std::make_shared<gdsl::GDSLPreprocessor>(name, program, makroDir);
        m_preprocessor->parsePpStatements();
        m_preprocessor->applyPpStatements();

        program = m_preprocessor->preprocessedProgram();

        std::cout << "Preprocessed program: " << std::endl << program << std::endl;

        if(m_preprocessor->name() != "")
            name = m_preprocessor->name();
    }
    catch(gdsl::PreprocessorException e)
    {
        setParseError("Preprocessor Error", e.what());
    }
}

void GDSLuiProgram::parse(string &name, string &program)
{
    try
    {
        m_program = std::make_shared<gdsl::GDSLProgram>(name, program);
    }
    catch(gdsl::ParseException e)
    {
        setParseError("Parsing Error", e.what());
    }
    catch(gdsl::OutOfBoundsException e)
    {
        setParseError("Out Of Bounds Error", e.what());
    }
    catch(GRBException e)
    {
        setParseError("Gurobi Error", e.getMessage());
    }
}

void GDSLuiProgram::preprocessAndParse(string &name, string &program, string& makroDir)
{
    preprocess(name, program, makroDir);

    if(m_parseFinishedWithError)
        return;

    parse(name, program);
}

std::vector< std::shared_ptr<GRBModel> > GDSLuiProgram::run(std::shared_ptr<gdsl::RandomAccessGraph> raGraph, std::shared_ptr<gdsl::Grid> grid, std::shared_ptr<GRBEnv> env)
{
    std::vector< std::shared_ptr<GRBModel> > models;
    m_numRuns = 0;

    while(true)
    {
        ++m_numRuns;
        std::cout << "Starting run " << m_numRuns << std::endl;

        try
        {
            auto constants = m_preprocessor->constants();
            m_program->setConstants(constants);

            models.push_back(runOneTime(raGraph, grid, env));

            std::vector<std::shared_ptr<GDSLVariableInfo> > varInfo;
            for(auto var : m_program->variables())
                varInfo.push_back(std::make_shared<GDSLVariableInfo>(var));

            m_varInfos[m_numRuns-1] = varInfo;
            m_finishedWithError.push_back(false);
        }
        catch(gdsl::PreprocessorException e)
        {
            setRunError(m_numRuns - 1, "Preprocessor Error", e.what());
        }
        catch(gdsl::ParseException e)
        {
            setRunError(m_numRuns - 1, "Parsing Error", e.what());
        }
        catch(gdsl::OutOfBoundsException e)
        {
            setRunError(m_numRuns - 1, "Out Of Bounds Error", e.what());
        }
        catch(GRBException e)
        {
            setRunError(m_numRuns - 1, "Gurobi Error", e.getMessage());
        }

        if(!m_preprocessor->next())
            break;
    }

    return models;
}

GDSLuiProgram::ResultMap GDSLuiProgram::variables() const
{
    return m_varInfos;
}

std::shared_ptr<gdsl::GDSLPreprocessor> GDSLuiProgram::preprocessor() const
{
    return m_preprocessor;
}

void GDSLuiProgram::setRunError(int run, string title, string message)
{
    m_errorTitle[run] = title;
    m_errorMessage[run] = message;
    m_finishedWithError.push_back(true);
}

void GDSLuiProgram::setParseError(string title, string message)
{
    m_parseErrorTitle = title;
    m_parseErrorMessage = message;
    m_parseFinishedWithError = true;
}

std::shared_ptr<GRBModel> GDSLuiProgram::runOneTime(std::shared_ptr<gdsl::RandomAccessGraph> raGraph, std::shared_ptr<gdsl::Grid> grid, std::shared_ptr<GRBEnv> env)
{
    m_program->runForGraph(raGraph, grid);

    std::shared_ptr<GRBModel> model = m_program->toGurobi(*env);
    model->optimize();

    return model;
}
