#include "gdslvariableinfo.hpp"

#include "../GDSL/parser/gdslvariable.hpp"
#include "../GDSL/parser/restrictions/gdslrestriction.hpp"
#include "../GDSL/parser/expressions/gdslprogram.hpp"

using namespace std::placeholders;

GDSLVariableInfo::GDSLVariableInfo(std::shared_ptr<gdsl::GDSLVariable> variable)
{
    m_defaultVar = gdsl::GDSLProgram::isDefaultVar(variable);
    m_restCount = variable->restrictions().size();
    m_ilpVarCount = variable->ilpVars().size();

    m_name = QString::fromStdString(variable->id());
    m_varString = QString::fromStdString(variable->toString());

    for(std::shared_ptr<gdsl::GDSLRestriction> rest : variable->restrictions())
    {
        m_restSizes.push_back(rest->length());
        m_restTypes.push_back(rest->outputType());
        m_restStrings.push_back(QString::fromStdString(rest->toString()));
    }

    if(!isDefaultVar())
    {
        int i = 0;
        for(std::shared_ptr<gdsl::IlpVariable> var : variable->ilpVars())
        {
            m_ilpValues[i] = QString::fromStdString(var->value());
            m_ilpRestStrings[i] = QString::fromStdString(variable->toString(i));

            ++i;
        }
    }

    computeDefAt(variable);
    computeDrawable(variable);
    computeSetMap(variable);
    if(!isDefaultVar())
        computeTree(variable);
}

int GDSLVariableInfo::restrictionCount() const
{
    return m_restCount;
}

int GDSLVariableInfo::restSize(int rest) const
{
    if(rest > restrictionCount() || rest < 0)
        return -1;

    return m_restSizes[rest];
}

int GDSLVariableInfo::ilpVarCount() const
{
    return m_ilpVarCount;
}

QString GDSLVariableInfo::restString(int index) const
{
    if(index > restrictionCount() || index < 0)
        return "";

    return m_restStrings[index];
}

QString GDSLVariableInfo::restStringForIlpVar(int index) const
{
    if(index > ilpVarCount() || index < 0)
        return "";

    return m_ilpRestStrings.at(index);
}

QString GDSLVariableInfo::ilpValueAt(int index) const
{
    if(index > ilpVarCount() || index < 0)
        return "";

    return QString("%1").arg(m_ilpValues.at(index));
}

QString GDSLVariableInfo::toString() const
{
    return m_varString;
}

QString GDSLVariableInfo::varName() const
{
    return m_name;
}

std::shared_ptr<TreeItem> GDSLVariableInfo::varToTree() const
{
    return m_tree;
}

bool GDSLVariableInfo::isDefaultVar() const
{
    return m_defaultVar;
}

bool GDSLVariableInfo::drawable() const
{
    return m_drawable;
}

bool GDSLVariableInfo::defAtRows() const
{
    return m_defAtRows;
}

bool GDSLVariableInfo::defAtCols() const
{
    return m_defAtCols;
}

bool GDSLVariableInfo::isNodeRest(int index) const
{
    if(index < restrictionCount() && m_restTypes[index] == gdsl::RestrictionType::NODEREST)
        return true;

    return false;
}

bool GDSLVariableInfo::isEdgeRest(int index) const
{
    if(index < restrictionCount() && m_restTypes[index] == gdsl::RestrictionType::EDGEREST)
        return true;

    return false;

}

bool GDSLVariableInfo::isObjectRest(int index) const
{
    if(index < restrictionCount() && m_restTypes[index] == gdsl::RestrictionType::OBJECTREST)
        return true;

    return false;

}

bool GDSLVariableInfo::isSetAt(int row, int column) const
{
    if(!m_drawable)
        return false;

    if(!m_defAtCols)
        column = 0;
    if(!m_defAtRows)
        row = 0;

    return (m_isSetMap.find(std::make_pair(row,column)) != m_isSetMap.end());
}

void GDSLVariableInfo::setColorForNodes(QString color)
{
    setAttributeForTreeType(color, m_tree, TreeItem::TreeItemType::NODE_ID, [](std::shared_ptr<TreeItem> item, QString color) {item->setColor(color);});
}

void GDSLVariableInfo::setColorForEdges(QString color)
{
    setAttributeForTreeType(color, m_tree, TreeItem::TreeItemType::EDGE_ID, [](std::shared_ptr<TreeItem> item, QString color) {item->setColor(color);});
}

void GDSLVariableInfo::computeSetMap(std::shared_ptr<gdsl::GDSLVariable> variable)
{
    if(!drawable())
        return;

    auto rests = variable->restrictions();
    gdsl::CombinationIterator it(rests);

    while(!it.isEnd())
    {
        auto singleRest = it.current();
        int row = 0;
        int column = 0;
        bool isSet = false;

        gdsl::GDSLVariable::IlpVarList ilpVars = variable->get(singleRest);
        for(std::shared_ptr<gdsl::IlpVariable> var : ilpVars)
            if(std::stoi(var->value()) == 1)
                isSet = true;

        if(isSet)
        {
            for(auto rest : singleRest)
            {
                if(rest->outputType() == gdsl::RestrictionType::ROWREST)
                    row = *rest->begin();
                else if(rest->outputType() == gdsl::RestrictionType::COLUMNREST)
                    column = *rest->begin();
            }

            std::pair<int,int> rowColumnPair(row, column);
            m_isSetMap.insert(std::make_pair(rowColumnPair, isSet));
        }

        it.next();
    }
}

void GDSLVariableInfo::computeDefAt(std::shared_ptr<gdsl::GDSLVariable> variable)
{
    for(auto rest : variable->restrictions())
    {       
        if(rest->outputType() == gdsl::RestrictionType::ROWREST)
            m_defAtRows = true;
        else if(rest->outputType() == gdsl::RestrictionType::COLUMNREST)
            m_defAtCols = true;
    }
}

void GDSLVariableInfo::computeTree(std::shared_ptr<gdsl::GDSLVariable> variable)
{
    m_tree = std::make_shared<TreeItem>();
    m_tree->setType(TreeItem::TreeItemType::ROOT);
    m_tree->setText(varName());

    for(int i = 0; i < ilpVarCount(); ++i)
    {
        std::shared_ptr<TreeItem> treePosition = m_tree;

        int column = 0;
        int row = 0;

        auto rests = variable->restFromId(i);

        for(auto it = rests->begin(); it < rests->end(); ++it)
        {
            gdsl::RestrictionType restType = (*it)->outputType();

            if(restType == gdsl::RestrictionType::ROWREST)
                row = *((*it)->begin());
            else if(restType == gdsl::RestrictionType::COLUMNREST)
                column = *((*it)->begin());
            else
            {
                auto item = findRootInTree(treePosition, *it);

                if(item != nullptr)
                    treePosition = item;
                else
                {
                    item = std::make_shared<TreeItem>(treePosition);
                    treePosition->addChild(item);

                    if(restType == gdsl::RestrictionType::NODEREST)
                    {
                        item->setText("nodes");
                        item->setType(TreeItem::TreeItemType::NODE_ROOT);
                    }
                    else if(restType == gdsl::RestrictionType::EDGEREST)
                    {
                        item->setText("edges");
                        item->setType(TreeItem::TreeItemType::EDGE_ROOT);
                    }
                }

                std::shared_ptr<TreeItem> dataItem = findDataInTree(item, *it);

                if(dataItem != nullptr)
                    treePosition = dataItem;

                else
                {
                    dataItem = std::make_shared<TreeItem>(item);
                    item->addChild(dataItem);

                    int objId = *((*it)->begin());

                    QString label;
                    QString origColor;

                    if(restType == gdsl::RestrictionType::NODEREST)
                    {
                        ogdf::node node = variable->attr()->nodeById(objId);
                        label = QString(variable->attr()->attr()->labelNode(node).cstr());
                        origColor = QString(variable->attr()->attr()->colorNode(node).cstr());
                        dataItem->setType(TreeItem::TreeItemType::NODE_ID);
                    }
                    else if(restType == gdsl::RestrictionType::EDGEREST)
                    {
                        ogdf::edge edge = variable->attr()->edgeById(objId);
                        label = QString(variable->attr()->attr()->labelEdge(edge).cstr());
                        origColor = QString(variable->attr()->attr()->colorEdge(edge).cstr());
                        dataItem->setType(TreeItem::TreeItemType::EDGE_ID);
                    }

                    dataItem->setLabel(label);
                    dataItem->setOrigColor(origColor);

                    std::string text = std::to_string(objId);
                    dataItem->setText(QString::fromStdString(text));

                    treePosition = dataItem;
                }
            }
        }

        std::shared_ptr<TreeItem> coordItem = std::make_shared<TreeItem>(treePosition);
        coordItem->setType(TreeItem::TreeItemType::COORD);
        coordItem->setCoords(row, column);
        treePosition->addChild(coordItem);
        coordItem->setText(QString("%1,%2").arg(row).arg(column));
        coordItem->setData(ilpValueAt(i));

        if(coordItem->data().toDouble() == 0.0)
            coordItem->setVisible(false);
    }
}

std::shared_ptr<TreeItem> GDSLVariableInfo::findRootInTree(std::shared_ptr<TreeItem> tree, std::shared_ptr<gdsl::GDSLRestriction> rest)
{
    for(auto it = tree->childrenBegin(); it < tree->childrenEnd(); ++it)
    {
        if(rest->outputType() == gdsl::RestrictionType::NODEREST && (*it)->type() == TreeItem::TreeItemType::NODE_ROOT)
            return *it;
        else if(rest->outputType() == gdsl::RestrictionType::EDGEREST && (*it)->type() == TreeItem::TreeItemType::EDGE_ROOT)
            return *it;
    }

    return nullptr;
}

std::shared_ptr<TreeItem> GDSLVariableInfo::findDataInTree(std::shared_ptr<TreeItem> tree, std::shared_ptr<gdsl::GDSLRestriction> rest)
{
    QString data = QString("%1").arg(*rest->begin());

    for(auto it = tree->childrenBegin(); it < tree->childrenEnd(); ++it)
        if((*it)->text() == data)
            return *it;

    return nullptr;
}

void GDSLVariableInfo::setAttributeForTreeType(QString attribute, std::shared_ptr<TreeItem> item,
                                               TreeItem::TreeItemType type, std::function<void (std::shared_ptr<TreeItem>, QString)> setFunc)
{
    if(item->type() == type)
        setFunc(item, attribute);

    for(auto it = item->childrenBegin(); it < item->childrenEnd(); ++it)
        setAttributeForTreeType(attribute, *it, type, setFunc);
}

void GDSLVariableInfo::computeDrawable(std::shared_ptr<gdsl::GDSLVariable> variable)
{
    bool coordinateDefined = m_defAtCols || m_defAtRows;
    bool objectDefined = false;

    for(auto rest : variable->restrictions())
        if(rest->outputType() == gdsl::RestrictionType::NODEREST ||
                rest->outputType() == gdsl::RestrictionType::EDGEREST ||
                rest->outputType() == gdsl::RestrictionType::OBJECTREST)
            objectDefined = true;

    if(coordinateDefined && objectDefined)
        m_drawable = true;
}


void TreeItem::print(int i) const
{
    std::cout << std::string(i, ' ') << "(" << m_text.toStdString() << "[" << m_data.toStdString() << "]" << ")" << std::endl;
    ++i;
    for(auto child : m_children)
        child->print(i);
}
