#ifndef GDSLVARIABLEINFO_HPP
#define GDSLVARIABLEINFO_HPP

#include <memory>
#include <unordered_map>
#include <utility>
#include <queue>
#include <vector>

#include <QString>
#include <QPair>

#include <iostream>
#include <functional>

namespace gdsl
{
    class GDSLVariable;
    class GDSLRestriction;
    class GDSLRowRest;
    class GDSLColumnRest;
    enum class RestrictionType;
}

template <class T>
inline void hash_combine(std::size_t & seed, const T & v)
{
  std::hash<T> hasher;
  seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

namespace std
{
  template<typename S, typename T> struct hash<pair<S, T>>
  {
    inline size_t operator()(const pair<S, T> & v) const
    {
      size_t seed = 0;
      ::hash_combine(seed, v.first);
      ::hash_combine(seed, v.second);
      return seed;
    }
  };
}

class TreeItem
{
public:
    enum class TreeItemType {STANDARD, ROOT, NODE_ROOT, EDGE_ROOT, NODE_ID, EDGE_ID, COORD};

    TreeItem(std::shared_ptr<TreeItem> parent = nullptr) : m_coords(0,0) {m_parent = parent;}
    void setText(QString text) {m_text = text;}
    void setData(QString data) {m_data = data;}
    void setLabel(QString label) {m_label = label;}
    void setOrigColor(QString color) {m_origColor = color;}
    void setColor(QString color) {m_color = color;}
    void setType(TreeItemType type) {m_type = type;}
    void setVisible(bool visible) {m_visible = visible;}
    void setCoords(int x, int y) {m_coords = QPair<int,int>(x, y);}

    void addChild(std::shared_ptr<TreeItem> child) {m_children.push_back(child);}

    QString text() const {return m_text;}
    QString data() const {return m_data;}
    QString label() const {return m_label;}
    QString origColor() const {return m_origColor;}
    QString color() const {return m_color;}
    bool visible() const {return m_visible;}

    std::vector< std::shared_ptr<TreeItem> >::iterator childrenBegin() {return m_children.begin();}
    std::vector< std::shared_ptr<TreeItem> >::iterator childrenEnd() {return m_children.end();}
    int childCount() {return childrenEnd() - childrenBegin();}
    QPair<int, int> coords() const {return m_coords;}

    std::shared_ptr<TreeItem> parent() {return m_parent;}

    void print(int i) const;

    bool isLeaf() const {return m_children.empty();}

    TreeItemType type() const {return m_type;}

private:
    std::shared_ptr<TreeItem> m_parent;
    std::vector< std::shared_ptr<TreeItem> > m_children;

    TreeItemType m_type;
    bool m_visible = true;

    QString m_text = "";
    QString m_data = "";
    QString m_label = "";
    QString m_origColor = "#000000";
    QString m_color = "";
    QPair<int,int> m_coords;
};

class GDSLVariableInfo
{
public:
    GDSLVariableInfo(std::shared_ptr<gdsl::GDSLVariable> variable);

    int restrictionCount() const;
    int restSize(int rest) const;
    int ilpVarCount() const;

    QString restString(int index) const;
    QString restStringForIlpVar(int index) const;
    QString ilpValueAt(int index) const;

    QString toString() const;
    QString varName() const;

    std::shared_ptr<TreeItem> varToTree() const;

    bool isDefaultVar() const;
    bool drawable() const;
    bool defAtRows() const;
    bool defAtCols() const;
    bool isNodeRest(int index) const;
    bool isEdgeRest(int index) const;
    bool isObjectRest(int index) const;
    bool isSetAt(int row, int column) const;

    void setColorForNodes(QString color);
    void setColorForEdges(QString color);

private:
    void computeSetMap(std::shared_ptr<gdsl::GDSLVariable> variable);
    void computeDrawable(std::shared_ptr<gdsl::GDSLVariable> variable);
    void computeDefAt(std::shared_ptr<gdsl::GDSLVariable> variable);
    void computeTree(std::shared_ptr<gdsl::GDSLVariable> variable);

    std::shared_ptr<TreeItem> findRootInTree(std::shared_ptr<TreeItem> tree, std::shared_ptr<gdsl::GDSLRestriction> rest);
    std::shared_ptr<TreeItem> findDataInTree(std::shared_ptr<TreeItem> tree, std::shared_ptr<gdsl::GDSLRestriction> rest);

    void setAttributeForTreeType(QString attribute, std::shared_ptr<TreeItem> item, TreeItem::TreeItemType type,
                                 std::function<void (std::shared_ptr<TreeItem>, QString)> setFunc);

    std::unordered_map< std::pair<int,int>, bool > m_isSetMap;
    std::shared_ptr<TreeItem> m_tree;

    bool m_drawable = false;
    bool m_defAtRows = false;
    bool m_defAtCols = false;
    bool m_defaultVar = false;

    int m_restCount;
    std::vector<int> m_restSizes;
    std::vector<gdsl::RestrictionType> m_restTypes;
    std::vector<QString> m_restStrings;

    int m_ilpVarCount;
    std::vector<QString> m_ilpStrings;
    std::unordered_map<int, QString> m_ilpValues;
    std::unordered_map<int, QString> m_ilpRestStrings;

    QString m_name;
    QString m_varString;
};

#endif // GDSLVARIABLEINFO_HPP
