#include "gdslvariablestable.hpp"
#include "ui_gdslvariablestable.h"

GDSLVariablesTable::GDSLVariablesTable(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::GDSLVariablesTable)
{
    ui->setupUi(this);

    readSettings();

    ui->table->setDropIndicatorShown(true);
    ui->table->setDragEnabled(true);
    ui->table->setDragDropMode(QAbstractItemView::InternalMove);
    ui->table->viewport()->setAcceptDrops(true);

    connect(ui->table, SIGNAL(expanded(QModelIndex)), this, SLOT(expandedItem(QModelIndex)));
}

GDSLVariablesTable::~GDSLVariablesTable()
{
    delete ui;
}

void GDSLVariablesTable::addItem(std::shared_ptr<GDSLVariableInfo> info)
{
    if(!info->isDefaultVar() || m_showDefaults)
    {
        std::shared_ptr<TreeItem> tree = info->varToTree();
        auto root = new VariablesTableItem(tree);

        root->setCheckState(0, Qt::Checked);

        addTreeItem(root, tree);

        root->setFlags(root->flags() | Qt::ItemIsDragEnabled);

        ui->table->insertTopLevelItem(ui->table->topLevelItemCount(), root);
        connect(ui->table, SIGNAL(itemChanged(QTreeWidgetItem*,int)), this, SLOT(itemChangedSlot(QTreeWidgetItem*,int)));

        showOrHideCoordinates(root);
    }
}

void GDSLVariablesTable::clearTable()
{
    ui->table->clear();
}

QTreeWidgetItem *GDSLVariablesTable::headItem(int i)
{
    return ui->table->topLevelItem(i);
}

void GDSLVariablesTable::readSettings()
{
    QSettings settings;

    if(settings.value(SettingsKeys::VIEW_SHOW_COLORPICKER).toBool())
        ui->table->showColumn(COLOR_COLUMN);
    else
        ui->table->hideColumn(COLOR_COLUMN);

    if(settings.value(SettingsKeys::VIEW_SHOW_LABELS).toBool())
        ui->table->showColumn(LABEL_COLUMN);
    else
        ui->table->hideColumn(LABEL_COLUMN);

    auto newShowCoords = settings.value(SettingsKeys::VIEW_SHOW_COORDINATES).toBool();

    if(newShowCoords != m_showCoordinates)
    {
        m_showCoordinates = newShowCoords;
        showOrHideCoordinates();
    }

    auto newColors = settings.value(SettingsKeys::VIEW_SHOW_ORIGCOLORS).toBool();

    if(newColors != m_useOrigColors)
    {
        m_useOrigColors = newColors;
        setColorsEnabled();
        emit colorModeChanged(m_useOrigColors);
    }
}

void GDSLVariablesTable::showOrHideCoordinates()
{
    for(int i = 0; i < ui->table->topLevelItemCount(); ++i)
        showOrHideCoordinates(ui->table->topLevelItem(i));
}

void GDSLVariablesTable::showOrHideCoordinates(QTreeWidgetItem *item)
{
    if(item->childCount() == 0)
        item->setHidden(!m_showCoordinates);

    for(int i = 0; i < item->childCount(); ++i)
        showOrHideCoordinates(item->child(i));
}

void GDSLVariablesTable::setColorsEnabled()
{
    for(int i = 0; i < ui->table->topLevelItemCount(); ++i)
        setColorsEnabled(ui->table->topLevelItem(i));
}

void GDSLVariablesTable::setColorsEnabled(QTreeWidgetItem *item)
{
    auto widget = ui->table->itemWidget(item, COLOR_COLUMN);
    if(widget)
        widget->setEnabled(!m_useOrigColors);

    for(int i = 0; i < item->childCount(); ++i)
        setColorsEnabled(item->child(i));
}

void GDSLVariablesTable::addTreeItem(VariablesTableItem* root, std::shared_ptr<TreeItem> tree)
{
    root->setText(VARIABLE_COLUMN, tree->text());
    root->setText(DATA_COLUMN, tree->data());
    root->setText(LABEL_COLUMN, tree->label());
    root->setText(COLOR_COLUMN, "");

    if(tree->type() == TreeItem::TreeItemType::NODE_ID || tree->type() == TreeItem::TreeItemType::EDGE_ID)
    {
        auto button = new ButtonColorChooser(tree, m_useOrigColors);
        button->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Fixed);
        button->setFixedSize(32, 12);
        button->setAutoFillBackground(true);
        ui->table->setItemWidget(root, COLOR_COLUMN, button);

        connect(button, SIGNAL(colorChanged(std::shared_ptr<TreeItem>)), this, SLOT(colorChangedSlot(std::shared_ptr<TreeItem>)));
        connect(this, SIGNAL(colorModeChanged(bool)), button, SLOT(colorModeChanged(bool)));

        button->setEnabled(!m_useOrigColors);
    }

    root->setFlags(root->flags() | Qt::ItemIsUserCheckable);
    root->setFlags(root->flags() & ~Qt::ItemIsDropEnabled);
    root->setFlags(root->flags() & ~Qt::ItemIsDragEnabled);

    if(root->parent() && root->parent()->checkState(0) == Qt::Checked)
        root->setCheckState(0, Qt::Checked);

    for(auto it = tree->childrenBegin(); it < tree->childrenEnd(); ++it)
    {
        if(((*it)->type() != TreeItem::TreeItemType::COORD) || ((*it)->type() == TreeItem::TreeItemType::COORD && (*it)->data().toDouble() != 0.0))
        {
            auto item = new VariablesTableItem(root, *it);
            root->addChild(item);
            addTreeItem(item, *it);
        }
    }
}

void GDSLVariablesTable::toggleZeros(QTreeWidgetItem *item)
{
    for(int j = 0; j < item->childCount(); ++j)
    {
        QTreeWidgetItem* child = item->child(j);

        if(child->text(1).toDouble() == 0.0 && child->text(1) != "")
            child->setHidden(!m_showZeros);

        toggleZeros(child);
    }
}

void GDSLVariablesTable::updateCheckState(VariablesTableItem* item, Qt::CheckState checkState)
{
    for(int i = 0; i < item->childCount(); ++i)
    {
        item->child(i)->setCheckState(0, checkState);

        if(checkState == Qt::Checked)
            item->childTableItem(i)->setTreeItemVisibility(true);
        else
            item->childTableItem(i)->setTreeItemVisibility(false);

        updateCheckState(item->childTableItem(i), checkState);
    }
}

void GDSLVariablesTable::updateParentCheckState(VariablesTableItem* item)
{
    if(item->parent())
    {
        auto parent = item->parentTableItem();

        int checkedCount = 0;
        int partCheckCount = 0;
        int uncheckedCount = 0;
        for(int i = 0; i < item->parent()->childCount(); ++i)
        {
            auto siblingState = item->parent()->child(i)->checkState(0);
            if(siblingState == Qt::Checked)
                ++checkedCount;
            else if(siblingState == Qt::PartiallyChecked)
                ++partCheckCount;
            else if(siblingState == Qt::Unchecked)
                ++uncheckedCount;
        }

        if(checkedCount == item->parent()->childCount())
        {
            parent->setCheckState(0, Qt::Checked);
            parent->setTreeItemVisibility(true);
        }
        else if(uncheckedCount == item->parent()->childCount())
        {
            parent->setCheckState(0, Qt::Unchecked);
            parent->setTreeItemVisibility(false);
        }
        else if(partCheckCount > 0 || uncheckedCount > 0)
            parent->setCheckState(0, Qt::PartiallyChecked);
        else
            std::cout << "WTF" << std::endl;

        updateParentCheckState(parent);
    }
}

void GDSLVariablesTable::itemChangedSlot(QTreeWidgetItem *item, int column)
{
    auto treeItem = dynamic_cast<VariablesTableItem*>(item);

    ui->table->blockSignals(true);

    if(column != 0)
        return;

    Qt::CheckState checkState = item->checkState(0);

    if(checkState == Qt::Checked)
        treeItem->gdslTreeItem()->setVisible(true);
    else
        treeItem->gdslTreeItem()->setVisible(false);

    updateCheckState(treeItem, checkState);
    updateParentCheckState(treeItem);

    ui->table->blockSignals(false);

    emit checkStatesChanged(treeItem->gdslTreeItem());
}

void GDSLVariablesTable::colorChangedSlot(std::shared_ptr<TreeItem> item)
{
    emit colorChanged(item);
}

//TODO: Emit perhaps?
void GDSLVariablesTable::toggleDefaults(bool state)
{
    m_showDefaults = state;
}

void GDSLVariablesTable::toggleZeros(bool)
{
//    m_showZeros = state;

//    for(int i = 0; i < ui->table->topLevelItemCount(); ++i)
//    {
//        QTreeWidgetItem* topLevelItem = ui->table->topLevelItem(i);
//        toggleZeros(topLevelItem);
    //    }
}

void GDSLVariablesTable::expandedItem(QModelIndex)
{
    for(int i = 0; i < ui->table->columnCount(); ++i)
        ui->table->resizeColumnToContents(i);
}
