#include "gdslvarlayer.hpp"

GDSLVarLayer::GDSLVarLayer(std::shared_ptr<GDSLVariableInfo> var, QGridItem *grid, QBrush brush) :
    QGraphicsObject(),
    m_variable(var)
{
    m_layerChanged = true;
    m_visible = true;
    m_rows = grid->rows();
    m_columns = grid->columns();
    m_rect = grid->rectSize();
    m_brush = brush;

    readSettings();
}

void GDSLVarLayer::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    if(m_layerChanged && m_visible)
    {
        std::shared_ptr<TreeItem> tree = m_variable->varToTree();

        if(tree->type() == TreeItem::TreeItemType::EDGE_ROOT)
            drawTree(tree, painter, option, true, widget);
        else
            drawTree(tree, painter, option, false, widget);

        m_layerChanged = false;
        update(boundingRect());
    }
}

void GDSLVarLayer::readSettings()
{
    QSettings settings;

    m_useOrigColors = settings.value(SettingsKeys::VIEW_SHOW_ORIGCOLORS).toBool();

    for(auto child : childItems())
    {
        auto rectItem = dynamic_cast<GDSLRectItem*>(child);
        rectItem->setUseOrigColor(m_useOrigColors);
    }

    colorChanged(m_variable->varToTree());
}

void GDSLVarLayer::drawTree(std::shared_ptr<TreeItem> item, QPainter *painter, const QStyleOptionGraphicsItem *option, bool edgeSubTree, QWidget *widget)
{
    if(item->childCount() == 0)
    {
        if(item->type() == TreeItem::TreeItemType::COORD && item->data().toDouble() != 0.0)
        {
            auto coord = item->text();
            auto coords = coord.split(",");
            int row = coords[0].toInt();
            int column = coords[1].toInt();

            if(!m_variable->defAtRows() && m_rows != 1)
                for(int i = 0; i < m_rows; ++i)
                    insertRowItem(i, column, item, painter, option, edgeSubTree, widget);

            else if(!m_variable->defAtCols() && m_columns != 1)
                for(int i = 0; i < m_columns; ++i)
                    insertColItem(row, i, item, painter, option, edgeSubTree, widget);

            else
                insertNode(row, column, item, painter, option, edgeSubTree, widget);
        }
    }
    else
        for(auto it = item->childrenBegin(); it < item->childrenEnd(); ++it)
        {
            if(edgeSubTree || item->type() == TreeItem::TreeItemType::EDGE_ROOT)
                drawTree(*it, painter, option, true, widget);
            else
                drawTree(*it, painter, option, false, widget);
        }
}

void GDSLVarLayer::insertNode(int row, int column, std::shared_ptr<TreeItem> item, QPainter *painter,
                              const QStyleOptionGraphicsItem *option, bool edgeSubTree, QWidget *widget)
{
    GDSLRectItem* rect = new GDSLRectItem(item, m_rect, this);

    if(edgeSubTree)
        rect->setZValue(0);
    else
        rect->setZValue(1);

    rect->createRectItems();
    rect->moveBy(column * m_rect.width(),row * m_rect.height());
    rect->setBrush(m_brush);
    rect->paint(painter, option, widget);
    rect->setUseOrigColor(m_useOrigColors);
    rect->checkIfVisible();
}

void GDSLVarLayer::insertColItem(int row, int column, std::shared_ptr<TreeItem> item, QPainter *painter,
                                 const QStyleOptionGraphicsItem *option, bool edgeSubTree, QWidget *widget)
{
    GDSLColumnItem* rect = new GDSLColumnItem(item, m_rect, this);

    if(edgeSubTree)
        rect->setZValue(0);
    else
        rect->setZValue(1);

    rect->createRectItems();
    rect->moveBy(column * m_rect.width(),row * m_rect.height());
    rect->setBrush(m_brush);
    rect->paint(painter, option, widget);
    rect->setUseOrigColor(m_useOrigColors);
    rect->checkIfVisible();
}

void GDSLVarLayer::insertRowItem(int row, int column, std::shared_ptr<TreeItem> item, QPainter *painter,
                                 const QStyleOptionGraphicsItem *option, bool edgeSubTree, QWidget *widget)
{
    GDSLRowItem* rect = new GDSLRowItem(item, m_rect, this);

    if(edgeSubTree)
        rect->setZValue(0);
    else
        rect->setZValue(1);

    rect->createRectItems();
    rect->moveBy(column * m_rect.width(),row * m_rect.height());
    rect->setBrush(m_brush);
    rect->paint(painter, option, widget);
    rect->setUseOrigColor(m_useOrigColors);
    rect->checkIfVisible();
}

void GDSLVarLayer::toggleVisible(bool visible)
{
    for(auto item : childItems())
        item->setVisible(visible);
}

void GDSLVarLayer::checkStateChanged(std::shared_ptr<TreeItem> item)
{
    auto root = item;
    while(root->parent())
        root = root->parent();

    if(root->text() == m_variable->varName())
    {
        for(auto child : childItems())
        {
            auto rectItem = dynamic_cast<GDSLRectItem*>(child);
            rectItem->checkIfVisible();
        }
    }
}

void GDSLVarLayer::colorChanged(std::shared_ptr<TreeItem> item)
{
    auto root = item;
    while(root->parent())
        root = root->parent();

    if(root->text() == m_variable->varName())
    {
        for(auto child : childItems())
        {
            auto rectItem = dynamic_cast<GDSLRectItem*>(child);
            rectItem->checkColor();
        }
    }
}
