#include "gmlfileplot.hpp"
#include "ui_gmlfileplot.h"

#include <QDebug>

using namespace ogdf;

GmlFilePlot::GmlFilePlot(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::GmlFilePlot)
{
    ui->setupUi(this);
}

GmlFilePlot::~GmlFilePlot()
{
    delete ui;
}

void GmlFilePlot::plotFile(QString &path)
{
    Graph g;
    GraphAttributes ga(g, GraphAttributes::nodeGraphics |
                       GraphAttributes::edgeGraphics |
                       GraphAttributes::nodeLabel |
                       GraphAttributes::edgeLabel |
                       GraphAttributes::nodeColor |
                       GraphAttributes::edgeColor);

    ga.readGML(g, path.toStdString().c_str());
    plotGraph(g, ga);
}

void GmlFilePlot::plotGraph(std::shared_ptr<GraphAttributes> ga)
{
    plotGraph(ga->constGraph(), *ga);
}

void GmlFilePlot::plotGraph(const Graph &g, GraphAttributes &ga)
{
    ga.setAllWidth(10);
    ga.setAllHeight(10);

    if(g.numberOfNodes() == 0)
        return;

    layoutGraph(g, ga);

    QGraphicsScene* scene = new QGraphicsScene(this);
    ui->plot->setScene(scene);

    QMap<ogdf::node, QGraphicsEllipseItem*> nodeItemMap;

    auto group = new QGraphicsItemGroup;
    double n_x = 0;
    double n_y = 0;
    ogdf::node n;
    forall_nodes(n, g)
    {
        n_x = ga.x(n);
        n_y = ga.y(n);

        QGraphicsEllipseItem* item = drawNode(n_x, n_y, ga.width(n), ga.height(n), QColor(ga.colorNode(n).cstr()));
        nodeItemMap.insert(n, item);
        group->addToGroup(item);
    }

    ogdf::edge e;
    forall_edges(e, g)
    {
        QGraphicsEllipseItem* u = nodeItemMap.value(e->source());
        QGraphicsEllipseItem* v = nodeItemMap.value(e->target());

        ogdf::DPolyline bends = ga.bends(e);
        auto item = drawEdge(u, v, bends, ga.colorEdge(e).cstr());
        group->addToGroup(item);
    }

    group->setScale(1.5);
    scene->addItem(group);

//    QString numberOfText("#%1: %2");
//    auto numNodes = ui->plot->scene()->addText(numberOfText.arg("Nodes").arg(g.numberOfNodes()));
//    auto numEdges = ui->plot->scene()->addText(numberOfText.arg("Edges").arg(g.numberOfEdges()));
//    numEdges->moveBy(0, numNodes->boundingRect().height());
//    ui->plot->setAlignment(Qt::AlignLeft | Qt::AlignTop);
}

void GmlFilePlot::layoutGraph(const ogdf::Graph&, ogdf::GraphAttributes &ga)
{
    PlanarizationLayout pl;

    FastPlanarSubgraph *ps = new FastPlanarSubgraph;
    ps->runs(100);
    VariableEmbeddingInserter *ves = new VariableEmbeddingInserter;
    ves->removeReinsert(EdgeInsertionModule::rrAll);
    pl.setSubgraph(ps);
    pl.setInserter(ves);

    EmbedderMinDepthMaxFaceLayers *emb = new EmbedderMinDepthMaxFaceLayers;
    pl.setEmbedder(emb);

    OrthoLayout *ol = new OrthoLayout;
    ol->separation(20.0);
    ol->cOverhang(0.4);
    ol->setOptions(2+4);
    pl.setPlanarLayouter(ol);

    pl.call(ga);
}

QGraphicsEllipseItem *GmlFilePlot::drawNode(int x, int y, double width, double height, QColor color)
{
    QGraphicsEllipseItem* item = ui->plot->scene()->addEllipse(-width/2, -height/2, width, height);
    item->setBrush(QBrush(color));

    item->setPos(x, y);
    item->setZValue(1);
    return item;
}

QGraphicsPathItem* GmlFilePlot::drawEdge(QGraphicsEllipseItem *u, QGraphicsEllipseItem *v, ogdf::DPolyline &bends, QColor color)
{
    QPointF uPos = u->pos();
    QPointF vPos = v->pos();
    QPainterPath path;
    QPen pen(color);

    path.moveTo(uPos);

    for(auto it = bends.begin(); it != bends.end(); ++it)
        path.lineTo((*it).m_x, (*it).m_y);

    path.lineTo(vPos);

    return ui->plot->scene()->addPath(path, pen);
}
