#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "../GDSL/parser/grid.hpp"
#include "../GDSL/parser/randomaccessgraph.hpp"

#include "gurobi_c++.h"

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    checkSettings();

    ui->viewEditorSplitter->restoreState(m_settings.value(SettingsKeys::SPLITTER_VIEWEDITOR, ui->viewEditorSplitter->saveState()).toByteArray());
    ui->makroChooser->loadMakroList();

    m_highlighter = new GDSLHighlighter(ui->editor->document());

    const int tabStop = 4;
    QFontMetrics metrics(ui->editor->font());
    ui->editor->setTabStopWidth(tabStop * metrics.width(' '));

    m_lastGraphDirectory = QDir(m_settings.value(SettingsKeys::GEN_GRAPH_PATH).toString());
    m_lastProgramDirectory = QDir(m_settings.value(SettingsKeys::GEN_PROGRAM_PATH).toString());
    m_lastExportDirectory = QDir(QStandardPaths::writableLocation(QStandardPaths::HomeLocation));
}

MainWindow::~MainWindow()
{
    m_settings.setValue(SettingsKeys::SPLITTER_VIEWEDITOR, ui->viewEditorSplitter->saveState());

    delete m_highlighter;
    delete ui;
}

void MainWindow::on_actionSave_Program_triggered()
{
    QString saveFileName = QFileDialog::getSaveFileName(this, "Save Program as...", m_lastProgramDirectory.absolutePath(), "*.gdsl");

    if(saveFileName.isEmpty())
        return;

    QFileInfo fileInfo(saveFileName);
    m_lastProgramDirectory = fileInfo.absoluteDir();

    if(!fileInfo.fileName().endsWith(".gdsl"))
        saveFileName.append(".gdsl");

    QFile file(saveFileName);

    if(!file.open(QIODevice::WriteOnly))
        return;

    QTextStream out(&file);

    QString program = QString::fromStdString(getProgramString());
    out << program;
    out.flush();

    file.close();
}

void MainWindow::on_actionLoad_Program_triggered()
{
    QString openFileName = QFileDialog::getOpenFileName(this, "Open Program ...", m_lastProgramDirectory.absolutePath(), "*.gdsl");

    if(openFileName.isEmpty())
        return;

    QFileInfo fileInfo(openFileName);
    m_lastProgramDirectory = fileInfo.absoluteDir();

    QFile file(openFileName);

    if(!file.open(QIODevice::ReadOnly))
        return;

    QString program = file.readAll();
    readPpStmts(program);
}

void MainWindow::on_actionLoad_Graph_triggered()
{
    QString openFileName = QFileDialog::getOpenFileName(this, "Open Graph ...", m_lastGraphDirectory.absolutePath(), "*.gml");

    if(openFileName.isEmpty())
        return;

    QFileInfo fileInfo(openFileName);
    m_lastGraphDirectory = fileInfo.absoluteDir();

    long attributes = ogdf::GraphAttributes::nodeGraphics | ogdf::GraphAttributes::edgeGraphics | ogdf::GraphAttributes::nodeColor
            | ogdf::GraphAttributes::edgeColor | ogdf::GraphAttributes::nodeLabel | ogdf::GraphAttributes::edgeLabel;

    m_graph = std::make_shared<ogdf::Graph>();
    m_attr = std::make_shared<ogdf::GraphAttributes>(*m_graph, attributes);

    m_attr->readGML(*m_graph, openFileName.toStdString().c_str());
    ui->plotter->plotGraph(m_attr);
}

void MainWindow::on_runButton_clicked()
{
    std::cout << "Starting parsing ..." << std::endl;
    ui->runButton->setEnabled(false);

    if(m_attr == nullptr)
    {
        QMessageBox::critical(this, "No Graph loaded", "Please load a graph before running the program!");
        ui->runButton->setEnabled(true);
        return;
    }

    auto programStr = getProgramString();

    std::cout << "Running program: " << std::endl << programStr << std::endl;

    parse(programStr);
}

void MainWindow::on_actionExport_to_PNG_triggered()
{
    QString saveFileName = QFileDialog::getSaveFileName(this, "Export to PNG ...", m_lastExportDirectory.absolutePath());

    if(saveFileName.isEmpty())
        return;

    if(!saveFileName.endsWith(".png"))
        saveFileName += ".png";

    QImage image(1024, 1024, QImage::Format_ARGB32_Premultiplied);
    exportTo(&image);
    image.save(saveFileName);
}

void MainWindow::on_actionExport_to_PDF_triggered()
{
    QString saveFileName = QFileDialog::getSaveFileName(this, "Export to PNG ...", m_lastExportDirectory.absolutePath());

    if(saveFileName.isEmpty())
        return;

    if(!saveFileName.endsWith(".pdf"))
        saveFileName += ".pdf";

    QPdfWriter pdfWriter(saveFileName);
    exportTo(&pdfWriter);
}

void MainWindow::exportTo(QPaintDevice *device)
{
    QPainter painter(device);

    painter.setRenderHint(QPainter::Antialiasing);
    ui->resultTabWidget->exportCurrentTab(&painter);
    painter.end();
}

void MainWindow::parseFinished()
{
    std::cout << "Parsing finished" << std::endl;

    auto grid = std::make_shared<gdsl::Grid>(ui->colSpinBox->value(), ui->rowSpinBox->value());
    std::cout << "Grid has size " << grid->toString() << std::endl;
    auto raGraph = std::make_shared<gdsl::RandomAccessGraph>(m_attr);

    QThread* thread = new QThread;
    m_runner = std::make_shared<GDSLRunWorker>(m_parser->parsedProgram(), raGraph, grid);
    m_runner->moveToThread(thread);
    connect(thread, SIGNAL(started()), m_runner.get(), SLOT(process()));
    connect(m_runner.get(), SIGNAL(finished()), thread, SLOT(quit()));
    connect(thread, SIGNAL(finished()), this, SLOT(runFinished()));
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
    thread->start();
}

void MainWindow::runFinished()
{
    std::cout << "Run finished" << std::endl;
    ui->resultTabWidget->clear();

    for(int index = 0; index < m_runner->numRuns(); ++index)
    {
        if(m_runner->runFinishedWithError(index))
            ui->resultTabWidget->addError(QString::fromStdString(m_runner->getError(index).second), tr("Result %1").arg(index + 1));

        else
        {
            auto result = m_runner->variableInfo(index);
            auto resultDisplay = new GDSLDisplayResult(result, ui->rowSpinBox->value(), ui->colSpinBox->value(), this);

            connect(this, SIGNAL(settingsChanged()), resultDisplay, SLOT(settingsChangedSlot()));

            ui->resultTabWidget->addResult(resultDisplay, tr("Result %1").arg(index + 1));
        }
    }

    ui->runButton->setEnabled(true);
    ui->settingsResultTab->setCurrentIndex(1);
}

void MainWindow::displayWorkerError()
{
    auto error = m_parser->getError();
    QString title = QString::fromStdString(error.first);
    QString message = QString::fromStdString(error.second);
    QMessageBox::critical(this, title, message);

    ui->runButton->setEnabled(true);
}

void MainWindow::on_actionSettings_triggered()
{
    auto settings = new SettingsDialog(this);
    int result = settings->exec();

    if(result == QDialog::Accepted)
        emit settingsChanged();
}

void MainWindow::checkSettings()
{
    if(!m_settings.contains(SettingsKeys::MAKRO_PATH))
        m_settings.setValue(SettingsKeys::MAKRO_PATH, QStandardPaths::writableLocation(QStandardPaths::DataLocation));

    if(!m_settings.contains(SettingsKeys::GEN_GRAPH_PATH))
        m_settings.setValue(SettingsKeys::GEN_GRAPH_PATH, QStandardPaths::writableLocation(QStandardPaths::HomeLocation));

    if(!m_settings.contains(SettingsKeys::GEN_PROGRAM_PATH))
        m_settings.setValue(SettingsKeys::GEN_PROGRAM_PATH, QStandardPaths::writableLocation(QStandardPaths::HomeLocation));

    if(!m_settings.contains(SettingsKeys::HELPSTYLE_PATH))
    {
        QDir helpStyleDir(QStandardPaths::writableLocation(QStandardPaths::DataLocation));
        m_settings.setValue(SettingsKeys::HELPSTYLE_PATH, helpStyleDir.absoluteFilePath("helpstyle.css"));
    }

    if(!m_settings.contains(SettingsKeys::VIEW_SHOW_COLORPICKER))
        m_settings.setValue(SettingsKeys::VIEW_SHOW_COLORPICKER, true);

    if(!m_settings.contains(SettingsKeys::VIEW_SHOW_COORDINATES))
        m_settings.setValue(SettingsKeys::VIEW_SHOW_COORDINATES, false);

    if(!m_settings.contains(SettingsKeys::VIEW_SHOW_LABELS))
        m_settings.setValue(SettingsKeys::VIEW_SHOW_LABELS, true);

    if(!m_settings.contains(SettingsKeys::VIEW_SHOW_ORIGCOLORS))
        m_settings.setValue(SettingsKeys::VIEW_SHOW_ORIGCOLORS, false);
}

std::string MainWindow::getProgramString()
{
    auto makros = ui->makroChooser->processedMakros();
    std::string programStr;

    for(auto makro : makros)
    {
        programStr.append(makro);
        programStr.append("\n");
    }

    auto constants = ui->constants->preProcessorStrings();

    for(auto constant : constants)
    {
        programStr.append(constant.toStdString());
        programStr.append("\n");
    }

    auto executeForString = ui->executeFor->preProcessorString();

    if(executeForString != "")
    {
        programStr.append(executeForString.toStdString());
        programStr.append("\n");
    }

    programStr.append(ui->editor->toPlainText().toStdString());
    programStr.append("\n");

    return programStr;
}

void MainWindow::parse(const std::string &programStr)
{
    QThread* thread = new QThread();
    auto makroDir = m_settings.value(SettingsKeys::MAKRO_PATH).toString().toStdString();
    m_parser = std::make_shared<GDSLParseWorker>(programStr, makroDir);
    m_parser->moveToThread(thread);
    connect(m_parser.get(), SIGNAL(error()), this, SLOT(displayWorkerError()));
    connect(thread, SIGNAL(started()), m_parser.get(), SLOT(process()));
    connect(m_parser.get(), SIGNAL(finished()), thread, SLOT(quit()));
    connect(thread, SIGNAL(finished()), this, SLOT(parseFinished()));
    connect(thread, SIGNAL(finished()), thread, SLOT(deleteLater()));
    thread->start();
}

void MainWindow::readPpStmts(const QString &fullProgram)
{
    auto programStr = fullProgram.toStdString();
    std::string name = "";
    auto makroDir =  m_settings.value(SettingsKeys::MAKRO_PATH).toString().toStdString();
    gdsl::GDSLPreprocessor preprocessor(name, programStr, makroDir);
    preprocessor.parsePpStatements();

    ui->constants->reset();
    ui->executeFor->reset();

    for(std::string intVar : preprocessor.intVars())
    {
        QString start = QString::fromStdString(preprocessor.startForVar(intVar));
        QString end = QString::fromStdString(preprocessor.endForVar(intVar));
        QString name = QString::fromStdString(intVar);

        if(start == end)
            ui->constants->addConstant(name, "int");
        else
            ui->executeFor->setExecFor(name, start.toInt(), end.toInt());
    }

    for(std::string boolVar : preprocessor.boolVars())
    {
        QString name = QString::fromStdString(boolVar);
        ui->constants->addConstant(name, "bool");
    }

    for(std::string cons : preprocessor.constVars())
    {
        QString value = QString::fromStdString(preprocessor.startForVar(cons));
        QString name = QString::fromStdString(cons);

        ui->constants->addConstant(name, value);
    }

    ui->makroChooser->loadFromDependencyGraph(preprocessor.dependencies());
    ui->editor->setPlainText(QString::fromStdString(preprocessor.programWithoutPpStmts()));
}
